"""
Copyright (C) 2019-2020 Cisco Systems, Inc. and/or its affiliates. All rights reserved.

This module provides an assortment of helper functions that Mussels depends on.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from collections import defaultdict, namedtuple
import platform

NVC = namedtuple("NVC", "name version cookbook")


def version_keys(s):
    """
    `key` function enabling python's `sort` function to sort version strings.
    """
    import re

    keys = []
    for u in s.split("."):
        for v in re.split(r"(\d+)", u):
            try:
                val = int(v)
            except:
                val = str(v)
            keys.append(val)
    return keys


def sort_cookbook_by_version(items) -> defaultdict:
    """
    Sort items, and determine the highest versions.
    """
    sorted_items: defaultdict = defaultdict(list)

    for item in items:
        versions_list = list(items[item].keys())
        versions_list.sort(key=version_keys)
        versions_list.reverse()
        for version in versions_list:
            sorted_items[item].append(version)

    return sorted_items


PLATFORMS = {
    "posix": [
        "linux",
        "darwin",
        "macos",
        "osx",
        "freebsd",
        "openbsd",
        "sunos",
        "aix",
        "hp-ux",
    ],
    "unix": ["darwin", "macos", "osx", "freebsd", "openbsd", "sunos", "aix", "hp-ux"],
}


def platform_matches(requested_platform: str, specific_platform) -> bool:
    """
    Compare two platforms.
    Common platforms:
    - Windows
    - macos / darwin / osx
    - linux
    - unix (macos, sunos, bsd unix's)
    - *nix / posix (not windows)
    :return: True  if current platform matches requested platform.
    :return: False otherwise.
    """
    specific_platform = specific_platform.lower()
    requested_platform = requested_platform.lower()

    if requested_platform == specific_platform:
        return True

    elif (
        requested_platform == "mac"
        or requested_platform == "macos"
        or requested_platform == "osx"
    ) and specific_platform == "darwin":
        return True

    if (requested_platform == "unix") and (
        specific_platform == "darwin"
        or specific_platform == "sunos"
        or "bsd" in specific_platform
    ):
        return True

    elif requested_platform == "*nix" or requested_platform == "posix":
        if specific_platform != "windows":
            return True

    else:
        return False

    return False


def platform_is(requested_platform: str) -> bool:
    """
    Compare requested platform with current platform.
    Common platforms:
    - Win / Windows
    - Mac / macOS / Darwin
    - Linux
    - Unix (Mac, SunOS, BSD unix's)
    - *nix / posix (Not Windows)
    :return: True  if current platform matches requested platform.
    :return: False otherwise.
    """
    return platform_matches(requested_platform, platform.system())


def pick_platform(requested_platform: str, platform_options: list) -> str:
    """
    Given a list of platforms, pick the one that most closely matches the current platform.
    Prefer exact, allow superset.
    :return: string name of selected platform.
    """
    if requested_platform in platform_options:
        return requested_platform

    for option in platform_options:
        if platform_matches(option, requested_platform):
            return option

    return ""


def compare_versions(version_a: str, version_b: str) -> int:
    """
    Evaluate version strings of two versions.
    Compare if version A against version B.
    :return: -1  if A < B
    :return: 0   if A == B
    :return: 1   if A > B
    """
    if version_a == version_b:
        return 0

    versions_list = [version_a, version_b]

    versions_list.sort(key=version_keys)

    if versions_list[0] == version_a:
        return -1
    else:
        return 1


def get_item_version(item_name: str, sorted_items: dict, target: str = "", logger = None) -> NVC:
    """
    Convert a item name in the below format to a (name, version) tuple:

        [cookbook:]name[>=,<=,>,<,(==|=|@)version]

    Examples:
        - meepioux
        - blarghus>=1.2.3
        - wheeple@0.2.0
        - pyplo==5.1.0g
        - scrapbook:sasquatch<2.0.0
        - scrapbook: minnow < 0.1.12

    The highest available version will be selected if one is not specified.
    Version requirements will whittle down the list of available versions
    in the sorted_items list.

    If a specific version is specified, all other versions will be disqualified (pruned).

    If no versions remain that satisfy build qualifications, an exception will be raised.

    :return: named tuple describing the highest qualified version:
        NVC(
            "name"->str,
            "version"->str,
            "cookbook"->str,
        )
    """

    def select_cookbook_version(nvc, item_version, target: str = "") -> bool:
        cookbook_selected = False

        def cookbook_has_build_target(cookbooks_item: dict, target) -> bool:
            if target == "":
                return True

            for each_platform in cookbooks_item:
                # Note: sorted_items has been filtered down to compatible platform.
                #       No need to check with platform_is()
                if target in cookbooks_item[each_platform]:
                    return True
            return False

        # Prefer local over all else, enabling monkey-patching of recipes.
        if "local" in item_version["cookbooks"] and cookbook_has_build_target(
            item_version["cookbooks"]["local"], target
        ):
            if nvc["cookbook"] != "" and nvc["cookbook"] != "local":
                if logger:
                    logger.debug(f"Overriding {nvc_str(nvc['name'], nvc['version'], nvc['cookbook'])} with {nvc_str(nvc['name'], item_version['version'], 'local')}")
            nvc["version"] = item_version["version"]
            nvc["cookbook"] = "local"
            cookbook_selected = True
        else:
            if nvc["cookbook"] == "":
                # Any cookbook will do.
                for cookbook in item_version["cookbooks"]:
                    if cookbook_has_build_target(
                        item_version["cookbooks"][cookbook], target
                    ):
                        nvc["version"] = item_version["version"]
                        nvc["cookbook"] = cookbook
                        cookbook_selected = True
                        break
            else:
                # Check for requested cookbook.
                for cookbook in item_version["cookbooks"]:
                    if cookbook == nvc["cookbook"] and cookbook_has_build_target(
                        item_version["cookbooks"][cookbook], target
                    ):
                        nvc["version"] = item_version["version"]
                        cookbook_selected = True
                        break

        # Remove all other cookbooks for this item version.
        if cookbook_selected:
            item_version["cookbooks"] = {
                nvc["cookbook"]: item_version["cookbooks"][nvc["cookbook"]]
            }
        return cookbook_selected

    nvc = {"name": "", "version": "", "cookbook": ""}

    requested_item = item_name
    item_selected = False

    # Identify cookbook name, if provided.
    if ":" in item_name:
        cookbook, item = item_name.split(":")
        nvc["cookbook"] = cookbook.strip()
        item_name = item.strip()

    if ">=" in item_name:
        # GTE requirement found.
        name, version = item_name.split(">=")
        nvc["name"] = name.strip()
        version = version.strip()
        for i, item_version in enumerate(sorted_items[name]):
            cmp = compare_versions(item_version["version"], version)
            if cmp >= 0:
                # Version is good.
                if item_selected != True:
                    item_selected = select_cookbook_version(nvc, item_version, target)
            else:
                # Version is too low. Remove it, and subsequent versions.
                if logger != None and len(sorted_items[name][:i]) < len(sorted_items[name]):
                    logger.debug(f"{name} limited to version: {', '.join([item['version'] for item in sorted_items[name][:i]])}")

                sorted_items[name] = sorted_items[name][:i]
                break

    elif ">" in item_name:
        # GT requirement found.
        name, version = item_name.split(">")
        nvc["name"] = name.strip()
        version = version.strip()
        for i, item_version in enumerate(sorted_items[name]):
            cmp = compare_versions(item_version["version"], version)
            if cmp > 0:
                # Version is good.
                if item_selected != True:
                    item_selected = select_cookbook_version(nvc, item_version, target)
            else:
                # Version is too low. Remove it, and subsequent versions.
                if logger != None and len(sorted_items[name][:i]) < len(sorted_items[name]):
                    logger.debug(f"{name} limited to version: {', '.join([item['version'] for item in sorted_items[name][:i]])}")

                sorted_items[name] = sorted_items[name][:i]
                break

    elif "<=" in item_name:
        # LTE requirement found.
        name, version = item_name.split("<=")
        nvc["name"] = name.strip()
        version = version.strip()

        pruned = False
        # First, prune down to highest tolerable version
        if len(sorted_items[name]) > 0:

            while (
                len(sorted_items[name]) > 0
                and compare_versions(sorted_items[name][0]["version"], version) > 0
            ):
                # Remove a version from the sorted_items.
                sorted_items[name].remove(sorted_items[name][0])
                pruned = True

        # Then, prune down to the highest version provided by a the requested cookbook
        if len(sorted_items[name]) > 0:
            while len(sorted_items[name]) > 0 and not item_selected:
                item_selected = select_cookbook_version(
                    nvc, sorted_items[name][0], target
                )

                if not item_selected:
                    # Remove a version from the sorted_items.
                    sorted_items[name].remove(sorted_items[name][0])
                    pruned = True

            if logger != None and pruned:
                logger.debug(f"{name} limited to version: {', '.join([item['version'] for item in sorted_items[name]])}")

    elif "<" in item_name:
        # LT requirement found.
        name, version = item_name.split("<")
        nvc["name"] = name.strip()
        version = version.strip()

        pruned = False
        # First, prune down to highest tolerable version
        if len(sorted_items[name]) > 0:

            while (
                len(sorted_items[name]) > 0
                and compare_versions(sorted_items[name][0]["version"], version) >= 0
            ):
                # Remove a version from the sorted_items.
                sorted_items[name].remove(sorted_items[name][0])
                pruned = True

        # Then, prune down to the highest version provided by a the requested cookbook
        if len(sorted_items[name]) > 0:
            while len(sorted_items[name]) > 0 and not item_selected:
                item_selected = select_cookbook_version(
                    nvc, sorted_items[name][0], target
                )

                if not item_selected:
                    # Remove a version from the sorted_items.
                    sorted_items[name].remove(sorted_items[name][0])
                    pruned = True

            if logger != None and pruned:
                logger.debug(f"{name} limited to version: {', '.join([item['version'] for item in sorted_items[name]])}")

    else:
        eq_cond = False
        if "==" in item_name:
            name, version = item_name.split("==")
            eq_cond = True
        elif "=" in item_name:
            name, version = item_name.split("=")
            eq_cond = True
        elif "-" in item_name:
            name, version = item_name.split("-")
            eq_cond = True
        elif "@" in item_name:
            name, version = item_name.split("@")
            eq_cond = True

        if eq_cond == True:
            nvc["name"] = name.strip()
            nvc["version"] = version.strip()
            # EQ requirement found.
            # Try to find the specific version, and remove all others.
            item_selected = False
            for item_version in sorted_items[nvc["name"]]:
                if version == item_version["version"]:
                    item_selected = select_cookbook_version(nvc, item_version, target)
                    if item_selected:
                        if logger != None and len(sorted_items[nvc["name"]]) > 1:
                            logger.debug(f"{name} limited to version: {version}")

                        sorted_items[nvc["name"]] = [item_version]
                        break

        else:
            # No version requirement found.
            nvc["name"] = item_name.strip()

            for item_version in sorted_items[nvc["name"]]:
                item_selected = select_cookbook_version(nvc, item_version, target)
                if item_selected:
                    break

    if not item_selected:
        if target == "":
            raise Exception(
                f"No versions available to satisfy requirement for {requested_item}.\nThe requested version may have been filtered out by requirements for another recipe."
            )
        else:
            raise Exception(
                f"No versions available to satisfy requirement for {requested_item} ({target}).\nThe requested version may have been filtered out by requirements for another recipe."
            )

    return NVC(nvc["name"], nvc["version"], nvc["cookbook"])


def nvc_str(name, version, cookbook: str = ""):
    def nv_str(name, version):
        if version != "":
            return f"{name}-{version}"
        else:
            return name

    if cookbook != "":
        return f"{cookbook}:{nv_str(name, version)}"
    else:
        return nv_str(name, version)




"""
Copyright (C) 2019-2020 Cisco Systems, Inc. and/or its affiliates. All rights reserved.

This module provides the base class for every Recipe.

This includes the logic for how to perform a build and how to relocate (or "install")
select files to a directory structure where other recipes can find & depend on them.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import datetime
from distutils import dir_util
import glob
import inspect
from io import StringIO
import logging
import os
import platform
import shutil
import stat
import subprocess
import sys
import tarfile
import time
import zipfile

import requests
import urllib.request
import patch

from mussels.utils.versions import pick_platform, nvc_str


class BaseRecipe(object):
    """
    Base class for Mussels recipe.
    """

    name = "sample"
    version = "1.2.3"

    # Set collection to True if this is just a collection of recipes to build, and not an actual recipe.
    # If True, only `dependencies` and `required_tools` matter. Everything else should be omited.
    is_collection = False

    url = "https://sample.com/sample.tar.gz"  # URL of project release materials.

    # archive_name_change is a tuple of strings to replace.
    # For example:
    #     ("v", "nghttp2-")
    # will change:
    #     v2.3.4  to  nghttp2-2.3.4.
    # This hack is necessary because archives with changed names will extract to their original directory name.
    archive_name_change: tuple = ("", "")

    platforms: dict = {}  # Dictionary of recipe instructions for each platform.

    builds: dict = {}  # Dictionary of build paths.

    module_file: str = ""

    variables: dict = {} # variables that will be evaluated in build scripts

    def __init__(
        self,
        toolchain: dict,
        platform: str,
        target: str,
        install_dir: str = "",
        data_dir: str = "",
        work_dir: str = "",
        log_dir: str = "",
        download_dir: str = "",
        log_level: str = "DEBUG",
    ):
        """
        Download the archive (if necessary) to the Downloads directory.
        Extract the archive to the temp directory so it is ready to build.
        """
        self.toolchain = toolchain
        self.platform = platform
        self.target = target

        if data_dir == "":
            # No temp dir provided, build in the current working directory.
            self.data_dir = os.getcwd()
        else:
            self.data_dir = os.path.abspath(data_dir)

        self.install_dir = install_dir

        if download_dir != "":
            self.download_dir = download_dir
        else:
            self.download_dir = os.path.join(self.data_dir, "cache", "downloads")

        if log_dir != "":
            self.log_dir = log_dir
        else:
            self.log_dir = os.path.join(self.data_dir, "logs", "recipes")

        if work_dir != "":
            self.work_dir = work_dir
        else:
            self.work_dir = os.path.join(self.data_dir, "cache", "work")

        self.module_dir = os.path.split(self.module_file)[0]

        if "patches" in self.platforms[self.platform][self.target]:
            self.patch_dir = os.path.join(
                self.module_dir, self.platforms[self.platform][self.target]["patches"]
            )
        else:
            self.patch_dir = ""

        self._init_logging(log_level)

    def _init_logging(self, level="DEBUG"):
        """
        Initializes the logging parameters
        """
        levels = {
            "DEBUG": logging.DEBUG,
            "INFO": logging.INFO,
            "WARN": logging.WARNING,
            "WARNING": logging.WARNING,
            "ERROR": logging.ERROR,
        }

        os.makedirs(self.log_dir, exist_ok=True)

        self.logger = logging.getLogger(f"{nvc_str(self.name, self.version)}")

        formatter = logging.Formatter(
            fmt="%(asctime)s - %(levelname)s:  %(message)s",
            datefmt="%m/%d/%Y %I:%M:%S %p",
        )

        self.log_file = os.path.join(
            self.log_dir,
            f"{nvc_str(self.name, self.version)}.{datetime.datetime.now()}.log".replace(
                ":", "_"
            ),
        )
        filehandler = logging.FileHandler(filename=self.log_file)
        filehandler.setLevel(logging.DEBUG)
        filehandler.setFormatter(formatter)

        self.logger.addHandler(filehandler)
        self.logger.setLevel(levels[os.environ.get("LOG_LEVEL", level)])

    def _download_archive(self) -> bool:
        """
        Use the URL to download the archive if it doesn't already exist in the Downloads directory.
        """
        os.makedirs(self.download_dir, exist_ok=True)

        # Determine download path from URL &  possible archive name change.
        self.archive = self.url.split("/")[-1]
        if self.archive_name_change[0] != "":
            self.archive = self.archive.replace(
                self.archive_name_change[0], self.archive_name_change[1]
            )
        self.download_path = os.path.join(
            self.download_dir, self.archive
        )

        # Exit early if we already have the archive.
        if os.path.exists(self.download_path):
            self.logger.debug(f"Archive already downloaded.")
            return True

        self.logger.info(f"Downloading {self.url}")
        self.logger.info(f"         to {self.download_path} ...")

        if self.url.startswith("ftp"):
            try:
                urllib.request.urlretrieve(self.url, self.download_path)
            except Exception as exc:
                self.logger.info(f"Failed to download archive from {self.url}, {exc}!")
                return False
        else:
            try:
                r = requests.get(self.url)
                with open(self.download_path, "wb") as f:
                    f.write(r.content)
            except Exception:
                self.logger.info(f"Failed to download archive from {self.url}!")
                return False

        return True

    def _extract_archive(self, rebuild: bool) -> bool:
        """
        Extract the archive found in Downloads directory, if necessary.
        """
        if self.archive.endswith(".tar.gz"):
            self.builds[self.target] = os.path.join(
                self.work_dir, self.target, f"{self.archive[:-len('.tar.gz')]}"
            )
        elif self.archive.endswith(".zip"):
            self.builds[self.target] = os.path.join(
                self.work_dir, self.target, f"{self.archive[:-len('.zip')]}"
            )
        elif self.archive.endswith(".tar.xz"):
            self.builds[self.target] = os.path.join(
                self.work_dir, self.target, f"{self.archive[:-len('.tar.xz')]}"
            )
        else:
            self.logger.error(
                f"Unexpected archive extension. Currently only supports .tar.gz and .zip!"
            )
            return False

        self.prior_build_exists = os.path.exists(self.builds[self.target])

        if self.prior_build_exists:
            if not rebuild:
                # Build directory already exists.  We're good.
                return True

            # Remove previous built, start over.
            self.logger.info(
                f"--rebuild: Removing previous {self.target} build directory:"
            )
            self.logger.info(f"   {self.builds[self.target]}")
            shutil.rmtree(self.builds[self.target])
            self.prior_build_exists = False

        os.makedirs(os.path.join(self.work_dir, self.target), exist_ok=True)

        # Make our own copy of the extracted source so we don't dirty the original.
        self.logger.debug(f"Preparing {self.target} build directory:")
        self.logger.debug(f"   {self.builds[self.target]}")

        if self.archive.endswith(".tar.gz"):
            # Un-tar
            self.logger.info(
                f"Extracting tarball archive {self.archive} to {self.builds[self.target]} ..."
            )

            tar = tarfile.open(self.download_path, "r:gz")
            tar.extractall(os.path.join(self.work_dir, self.target))
            tar.close()
        elif self.archive.endswith(".zip"):
            # Un-zip
            self.logger.info(
                f"Extracting zip archive {self.archive} to {self.builds[self.target]} ..."
            )

            zip_ref = zipfile.ZipFile(self.download_path, "r")
            zip_ref.extractall(os.path.join(self.work_dir, self.target))
            zip_ref.close()
        elif self.archive.endswith(".tar.xz"):
            # Un-tar
            self.logger.info(
                f"Extracting tarball archive {self.archive} to {self.builds[self.target]} ..."
            )

            tar = tarfile.open(self.download_path, "r:xz")
            tar.extractall(os.path.join(self.work_dir, self.target))
            tar.close()

        return True

    def _run_script(self, target, name, script) -> bool:
        """
        Run a script in the current working directory.
        """
        # Create a build script.
        if platform.system() == "Windows":
            script_name = f"_{name}.bat"
            newline = "\r\n"
        else:
            script_name = f"_{name}.sh"
            newline = "\n"

        with open(os.path.join(os.getcwd(), script_name), "w", newline=newline) as fd:
            # Evaluate "".format() syntax in the build script
            script = script.format(**self.variables)

            # Write the build commands to a file
            build_lines = script.splitlines()
            for line in build_lines:
                line = line.strip()
                if platform.system() == "Windows" and line.endswith("\\"):
                    fd.write(line.rstrip("\\") + " ")
                else:
                    fd.write(line + "\n")

        if platform.system() != "Windows":
            st = os.stat(script_name)
            os.chmod(script_name, st.st_mode | stat.S_IEXEC)

        # Run the build script.
        process = subprocess.Popen(
            os.path.join(os.getcwd(), script_name),
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
        with process.stdout:
            for line in iter(process.stdout.readline, b""):
                self.logger.debug(line.decode("utf-8", errors='replace').strip())
        process.wait()
        if process.returncode != 0:
            self.logger.warning(
                f"{nvc_str(self.name, self.version)} {target} build failed!"
            )
            self.logger.warning(f"Command:")
            for line in script.splitlines():
                self.logger.warning(line)
            self.logger.warning(f"Exit code: {process.returncode}")
            self.logger.error(f'"{name}" script failed for {target} build')
            return False

        return True

    def build(self, rebuild: bool = False) -> bool:
        """
        Patch source materials if not already patched.
        Then, for each architecture, run the build commands if the output files don't already exist.
        """
        if self.is_collection:
            self.logger.debug(
                f"Build completed for recipe collection {nvc_str(self.name, self.version)}"
            )
            return True

        os.makedirs(self.work_dir, exist_ok=True)

        # Download and build if necessary.
        if not self._download_archive():
            self.logger.error(
                f"Failed to download source archive for {nvc_str(self.name, self.version)}"
            )
            return False

        # Extract to the work_dir.
        if not self._extract_archive(rebuild):
            self.logger.error(
                f"Failed to extract source archive for {nvc_str(self.name, self.version)}"
            )
            return False

        if not os.path.isdir(self.patch_dir):
            self.logger.debug(f"No patch directory found.")
        else:
            # Patches exists for this recipe.
            self.logger.debug(
                f"Patch directory found for {nvc_str(self.name, self.version)}."
            )
            if not os.path.exists(
                os.path.join(self.builds[self.target], "_mussles.patched")
            ):
                # Not yet patched. Apply patches.
                self.logger.info(
                    f"Applying patches to {nvc_str(self.name, self.version)} ({self.target}) build directory ..."
                )
                for patchfile in os.listdir(self.patch_dir):
                    if patchfile.endswith(".diff") or patchfile.endswith(".patch"):
                        self.logger.info(f"Attempting to apply patch: {patchfile}")
                        pset = patch.fromfile(os.path.join(self.patch_dir, patchfile))
                        patched = pset.apply(1, root=self.builds[self.target])
                        if not patched:
                            self.logger.error(f"Patch failed!")
                            return False
                    else:
                        self.logger.info(
                            f"Copying new file {patchfile} to {nvc_str(self.name, self.version)} ({self.target}) build directory ..."
                        )
                        shutil.copyfile(
                            os.path.join(self.patch_dir, patchfile),
                            os.path.join(self.builds[self.target], patchfile),
                        )

                with open(
                    os.path.join(self.builds[self.target], "_mussles.patched"), "w"
                ) as patchmark:
                    patchmark.write("patched")

        build_scripts = self.platforms[self.platform][self.target]["build_script"]

        self.logger.info(
            f"Attempting to build {nvc_str(self.name, self.version)} for {self.target}"
        )

        self.variables["includes"] = os.path.join(self.install_dir, "include").replace("\\", "/")
        self.variables["libs"] = os.path.join(self.install_dir, "lib").replace("\\", "/")
        self.variables["install"] = os.path.join(self.install_dir).replace("\\", "/")
        self.variables["build"] = os.path.join(self.builds[self.target]).replace("\\", "/")
        self.variables["target"] = self.target

        for tool in self.toolchain:
            # Add each tool from the toolchain to the PATH environment variable.
            if self.toolchain[tool].tool_path != "":
                self.logger.debug(
                    f"Adding tool {tool} path {self.toolchain[tool].tool_path} to PATH"
                )
                os.environ["PATH"] = (
                    self.toolchain[tool].tool_path + os.pathsep + os.environ["PATH"]
                )

            # Collect tool variables for use in the build.
            platform_options = self.toolchain[tool].platforms.keys()
            matching_platform = pick_platform(platform.system(), platform_options)

            if "variables" in self.toolchain[tool].platforms[matching_platform]:
                # The following will allow us to format strings like "echo {tool.variable}"
                tool_vars = lambda: None
                for variable in self.toolchain[tool].platforms[matching_platform]["variables"]:
                    setattr(tool_vars, variable, self.toolchain[tool].platforms[matching_platform]["variables"][variable])
                self.variables[tool] = tool_vars

        cwd = os.getcwd()
        os.chdir(self.builds[self.target])

        if not self.prior_build_exists:
            # Run "configure" script, if exists.
            if "configure" in build_scripts.keys():
                if not self._run_script(
                    self.target, "configure", build_scripts["configure"]
                ):
                    self.logger.error(
                        f"{nvc_str(self.name, self.version)} {self.target} build failed."
                    )
                    os.chdir(cwd)
                    return False

        else:
            os.chdir(self.builds[self.target])

        # Run "make" script, if exists.
        if "make" in build_scripts.keys():
            if not self._run_script(self.target, "make", build_scripts["make"]):
                self.logger.error(
                    f"{nvc_str(self.name, self.version)} {self.target} build failed."
                )
                os.chdir(cwd)
                return False

        # Run "install" script, if exists.
        if "install" in build_scripts.keys():
            if not self._run_script(self.target, "install", build_scripts["install"]):
                self.logger.error(
                    f"{nvc_str(self.name, self.version)} {self.target} build failed."
                )
                os.chdir(cwd)
                return False

        self.logger.info(
            f"{nvc_str(self.name, self.version)} {self.target} build succeeded."
        )
        os.chdir(cwd)

        if not self._install():
            return False

        return True

    def _install(self):
        """
        Copy the headers and libs to an install directory.
        """
        os.makedirs(self.install_dir, exist_ok=True)

        self.logger.info(
            f"Copying {nvc_str(self.name, self.version)} install files to: {self.install_dir}."
        )

        if 'install_paths' not in self.platforms[self.platform][self.target]:
            self.logger.info(
                f"{nvc_str(self.name, self.version)} {self.target} nothing additional to install."
            )

        else:
            install_paths = self.platforms[self.platform][self.target]["install_paths"]

            for install_path in install_paths:

                for install_item in install_paths[install_path]:
                    item_installed = False
                    src_path = os.path.join(self.builds[self.target], install_item)

                    for src_filepath in glob.glob(src_path):
                        dst_path = os.path.join(
                            self.install_dir, install_path, os.path.basename(src_filepath)
                        )

                        # Remove prior installation, if exists.
                        if os.path.isdir(dst_path):
                            shutil.rmtree(dst_path)
                        elif os.path.isfile(dst_path):
                            os.remove(dst_path)

                        # Create the target install paths, if it doesn't already exist.
                        os.makedirs(os.path.split(dst_path)[0], exist_ok=True)

                        self.logger.debug(f"Copying: {src_filepath}")
                        self.logger.debug(f"     to: {dst_path}")

                        # Now copy the file or directory.
                        if os.path.isdir(src_filepath):
                            dir_util.copy_tree(src_filepath, dst_path)
                        else:
                            shutil.copyfile(src_filepath, dst_path)

                        item_installed = True

                    # Globbing only shows us files that actually exists.
                    # It's possible we didn't install anything at all.
                    # Verify that we installed at least one item.
                    if not item_installed:
                        self.logger.error(
                            f"Required target install files do not exist:  {src_path}"
                        )
                        return False

            self.logger.info(
                f"{nvc_str(self.name, self.version)} {self.target} install succeeded."
            )
        return True



#!/usr/bin/env python

r"""
  __    __     __  __     ______     ______     ______     __         ______
 /\ "-./  \   /\ \/\ \   /\  ___\   /\  ___\   /\  ___\   /\ \       /\  ___\
 \ \ \-./\ \  \ \ \_\ \  \ \___  \  \ \___  \  \ \  __\   \ \ \____  \ \___  \
  \ \_\ \ \_\  \ \_____\  \/\_____\  \/\_____\  \ \_____\  \ \_____\  \/\_____\
   \/_/  \/_/   \/_____/   \/_____/   \/_____/   \/_____/   \/_____/   \/_____/
"""

_description = """
A tool to download, build, and assemble application dependencies.
                                    Brought to you by the Clam AntiVirus Team.
"""
_copyright = """
Copyright (C) 2019-2020 Cisco Systems, Inc. and/or its affiliates. All rights reserved.
"""

"""
Author: Micah Snyder

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import os
import sys

import click
import coloredlogs
import importlib.metadata

from mussels.mussels import Mussels
from mussels.utils.click import MusselsModifier, ShortNames

logging.basicConfig()
module_logger = logging.getLogger("mussels")
coloredlogs.install(level="DEBUG", fmt="%(asctime)s %(name)s %(levelname)s %(message)s")
module_logger.setLevel(logging.DEBUG)

from colorama import Fore, Back, Style

#
# CLI Interface
#
@click.group(
    cls=MusselsModifier,
    epilog=Fore.BLUE
    + __doc__
    + Fore.GREEN
    + _description
    + f"\nVersion {importlib.metadata.version('mussels')}\n"
    + Style.RESET_ALL
    + _copyright,
)
def cli():
    pass


@cli.group(cls=ShortNames, help="Commands that operate on cookbooks.")
def cookbook():
    pass


@cookbook.command("list")
@click.option(
    "--verbose", "-V", is_flag=True, default=False, help="Verbose output. [optional]"
)
def cookbook_list(verbose: bool):
    """
    Print the list of all known cookbooks.
    """
    my_mussels = Mussels(load_all_recipes=True)

    my_mussels.list_cookbooks(verbose)


@cookbook.command("show")
@click.argument("cookbook", required=True)
@click.option(
    "--verbose", "-V", is_flag=True, default=False, help="Verbose output. [optional]"
)
def cookbook_show(cookbook: str, verbose: bool):
    """
    Show details about a specific cookbook.
    """
    my_mussels = Mussels(load_all_recipes=True)

    my_mussels.show_cookbook(cookbook, verbose)


@cookbook.command("update")
def cookbook_update():
    """
    Update the cookbooks from the internet.
    """
    my_mussels = Mussels(load_all_recipes=True)

    my_mussels.update_cookbooks()


@cookbook.command("trust")
@click.argument("cookbook", required=True)
@click.option(
    "--yes",
    "-y",
    is_flag=True,
    default=False,
    help="Confirm trust. [required for non-interactive modes]",
)
def cookbook_trust(cookbook, yes):
    """
    Trust a cookbook.
    """
    my_mussels = Mussels(load_all_recipes=True)

    if yes != True:
        print(
            f"\nDisclaimer: There is a non-zero risk when running code downloaded from the internet.\n"
        )
        response = input(
            f"Are you sure you would like to trust recipes from cookbook '{cookbook}'? [N/y] "
        )
        response = response.strip().lower()

        if response != "y" and response != "yes":
            return

    my_mussels.config_trust_cookbook(cookbook)


@cookbook.command("add")
@click.argument("cookbook", required=True)
@click.option("--author", "-a", default="", help="Author or company name")
@click.option("--url", "-u", default="", help="Git repository URL")
@click.option(
    "--trust",
    "-t",
    is_flag=True,
    default=False,
    help="Add as a trusted cookbook, enabling you to build directly from this cookbook.",
)
def cookbook_add(cookbook, author, url, trust):
    """
    Add a cookbook to the list of known cookbooks.
    """
    my_mussels = Mussels(load_all_recipes=True)

    my_mussels.config_add_cookbook(cookbook, author, url, trust=trust)


@cookbook.command("remove")
@click.argument("cookbook", required=True)
def cookbook_remove(cookbook):
    """
    Remove a cookbook from the list of known cookbooks.
    """
    my_mussels = Mussels(load_all_recipes=True)

    my_mussels.config_remove_cookbook(cookbook)


@cli.group(cls=ShortNames, help="Commands that operate on recipes.")
def recipe():
    pass


@recipe.command("list")
@click.option(
    "--verbose", "-V", is_flag=True, default=False, help="Verbose output. [optional]"
)
@click.option(
    "--all",
    "-a",
    is_flag=True,
    default=False,
    help="List all recipes, including those for other platforms. [optional]",
)
def recipe_list(verbose: bool, all: bool):
    """
    Print the list of all known recipes.
    An asterisk indicates default (highest) version.
    """
    my_mussels = Mussels(load_all_recipes=all)

    my_mussels.list_recipes(verbose)


@recipe.command("show")
@click.argument("recipe", required=True)
@click.option("--version", "-v", default="", help="Version. [optional]")
@click.option(
    "--verbose", "-V", is_flag=True, default=False, help="Verbose output. [optional]"
)
@click.option(
    "--all",
    "-a",
    is_flag=True,
    default=False,
    help="Show all recipe variants, including those for other platforms. [optional]",
)
def recipe_show(recipe: str, version: str, verbose: bool, all: bool):
    """
    Show details about a specific recipe.
    """
    my_mussels = Mussels(load_all_recipes=all)

    my_mussels.show_recipe(recipe, version, verbose)


@recipe.command("clone")
@click.argument("recipe", required=True)
@click.option(
    "--version", "-v", default="", help="Specific version to clone. [optional]"
)
@click.option(
    "--cookbook", "-c", default="", help="Specific cookbook to clone. [optional]"
)
@click.option("--dest", "-d", default="", help="Destination directory. [optional]")
def recipe_clone(recipe: str, version: str, cookbook: str, dest: str):
    """
    Copy a recipe to the current directory or to a specific directory.
    """
    my_mussels = Mussels(load_all_recipes=True)

    my_mussels.clone_recipe(recipe, version, cookbook, dest)


@recipe.command("build")
@click.argument("recipe", required=True)
@click.option(
    "--version",
    "-v",
    default="",
    help="Version of recipe to build. May not be combined with @version in recipe name. [optional]",
)
@click.option(
    "--cookbook", "-c", default="", help="Specific cookbook to use. [optional]"
)
@click.option("--target", "-t", default="", help="Target architecture. [optional]")
@click.option(
    "--dry-run",
    "-d",
    is_flag=True,
    help="Print out the version dependency graph without actually doing a build. [optional]",
)
@click.option(
    "--rebuild",
    "-r",
    is_flag=True,
    help="Re-build a recipe, even if already built. [optional]",
)
@click.option(
    "--install", "-i", default="", help="Install directory. [optional] Default is: ~/.mussels/install/<target>"
)
@click.option(
    "--work-dir", "-w", default="", help="Work directory. [optional] Default is: ~/.mussels/cache/work"
)
@click.option(
    "--log-dir", "-l", default="", help="Log directory. [optional] Default is: ~/.mussels/logs"
)
@click.option(
    "--download-dir", "-D", default="", help="Downloads directory. [optional] Default is: ~/.mussels/cache/downloads"
)
def recipe_build(
    recipe: str,
    version: str,
    cookbook: str,
    target: str,
    dry_run: bool,
    rebuild: bool,
    install: str,
    work_dir: str,
    log_dir: str,
    download_dir: str,
):
    """
    Download, extract, build, and install a recipe.
    """

    my_mussels = Mussels(
        install_dir=install,
        work_dir=work_dir,
        log_dir=log_dir,
        download_dir=download_dir,
    )

    results = []

    success = my_mussels.build_recipe(
        recipe, version, cookbook, target, results, dry_run, rebuild
    )
    if success == False:
        sys.exit(1)

    sys.exit(0)


@cli.group(cls=ShortNames, help="Commands that operate on tools.")
def tool():
    pass


@tool.command("list")
@click.option(
    "--verbose", "-V", is_flag=True, default=False, help="Verbose output. [optional]"
)
@click.option(
    "--all",
    "-a",
    is_flag=True,
    default=False,
    help="List all tools, including those for other platforms. [optional]",
)
def tool_list(verbose: bool, all: bool):
    """
    Print the list of all known tools.
    An asterisk indicates default (highest) version.
    """
    my_mussels = Mussels(load_all_recipes=all)

    my_mussels.list_tools(verbose)


@tool.command("show")
@click.argument("tool", required=True)
@click.option("--version", "-v", default="", help="Version. [optional]")
@click.option(
    "--verbose", "-V", is_flag=True, default=False, help="Verbose output. [optional]"
)
@click.option(
    "--all",
    "-a",
    is_flag=True,
    default=False,
    help="Show all tool variants, including those for other platforms. [optional]",
)
def tool_show(tool: str, version: str, verbose: bool, all: bool):
    """
    Show details about a specific tool.
    """
    my_mussels = Mussels(load_all_recipes=all)

    my_mussels.show_tool(tool, version, verbose)


@tool.command("clone")
@click.argument("tool", required=True)
@click.option(
    "--version", "-v", default="", help="Specific version to clone. [optional]"
)
@click.option(
    "--cookbook", "-c", default="", help="Specific cookbook to clone. [optional]"
)
@click.option("--dest", "-d", default="", help="Destination directory. [optional]")
def tool_clone(tool: str, version: str, cookbook: str, dest: str):
    """
    Copy a tool to the current directory or to a specific directory.
    """
    my_mussels = Mussels(load_all_recipes=True)

    my_mussels.clone_tool(tool, version, cookbook, dest)


@tool.command("check")
@click.argument("tool", required=False, default="")
@click.option(
    "--version",
    "-v",
    default="",
    help="Version of tool to check. May not be combined with @version in tool name. [optional]",
)
@click.option(
    "--cookbook", "-c", default="", help="Specific cookbook to use. [optional]"
)
def tool_check(tool: str, version: str, cookbook: str):
    """
    Check if a tool is installed.
    """

    my_mussels = Mussels()

    results = []

    success = my_mussels.check_tool(tool, version, cookbook, results)
    if success == False:
        sys.exit(1)

    sys.exit(0)


@cli.group(cls=ShortNames, help="Commands to clean up.")
def clean():
    pass


@clean.command("cache")
def clean_cache():
    """
    Clear the cache files.
    """
    my_mussels = Mussels(load_all_recipes=True)

    my_mussels.clean_cache()


@clean.command("install")
def clean_install():
    """
    Clear the install files.
    """
    my_mussels = Mussels(load_all_recipes=True)

    my_mussels.clean_install()


@clean.command("logs")
def clean_logs():
    """
    Clear the logs files.
    """
    my_mussels = Mussels(load_all_recipes=True)

    my_mussels.clean_logs()


@clean.command("all")
def clean_all():
    """
    Clear the all files.
    """
    my_mussels = Mussels(load_all_recipes=True)

    my_mussels.clean_all()


#
# Command Aliases
#
@cli.command("build")
@click.argument("recipe", required=True)
@click.option(
    "--version",
    "-v",
    default="",
    help="Version of recipe to build. May not be combined with @version in recipe name. [optional]",
)
@click.option(
    "--cookbook", "-c", default="", help="Specific cookbook to use. [optional]"
)
@click.option("--target", "-t", default="", help="Target architecture. [optional]")
@click.option(
    "--dry-run",
    "-d",
    is_flag=True,
    help="Print out the version dependency graph without actually doing a build. [optional]",
)
@click.option(
    "--rebuild",
    "-r",
    is_flag=True,
    help="Re-build a recipe, even if already built. [optional]",
)
@click.option(
    "--install", "-i", default="", help="Install directory. [optional] Default is: ~/.mussels/install/<target>"
)
@click.option(
    "--work-dir", "-w", default="", help="Work directory. [optional] Default is: ~/.mussels/cache/work"
)
@click.option(
    "--log-dir", "-l", default="", help="Log directory. [optional] Default is: ~/.mussels/logs"
)
@click.option(
    "--download-dir", "-D", default="", help="Downloads directory. [optional] Default is: ~/.mussels/cache/downloads"
)
@click.pass_context
def build_alias(
    ctx,
    recipe: str,
    version: str,
    cookbook: str,
    target: str,
    dry_run: bool,
    rebuild: bool,
    install: str,
    work_dir: str,
    log_dir: str,
    download_dir: str,
):
    """
    Download, extract, build, and install a recipe.

    This is just an alias for `recipe build`.
    """
    ctx.forward(recipe_build)


@cli.command("list")
@click.pass_context
@click.option(
    "--verbose", "-V", is_flag=True, default=False, help="Verbose output. [optional]"
)
@click.option(
    "--all",
    "-a",
    is_flag=True,
    default=False,
    help="List all recipes, including those for other platforms. [optional]",
)
def list_alias(ctx, verbose: bool, all: bool):
    """
    List a list of recipes you can build on this platform.

    This is just an alias for `recipe list`.
    """
    ctx.forward(recipe_list)


@cli.command("show")
@click.pass_context
@click.argument("recipe", required=True)
@click.option("--version", "-v", default="", help="Version. [optional]")
@click.option(
    "--verbose", "-V", is_flag=True, default=False, help="Verbose output. [optional]"
)
@click.option(
    "--all",
    "-a",
    is_flag=True,
    default=False,
    help="Show all recipe variants, including those for other platforms. [optional]",
)
def show_alias(ctx, recipe: str, version: str, verbose: bool, all: bool):
    """
    Show details about a specific recipe.

    This is just an alias for `recipe show`.
    """
    ctx.forward(recipe_show)


@cli.command("update")
@click.pass_context
def update_alias(ctx):
    """
    Update local copy of cookbooks (using Git).

    This is just an alias for `recipe show`.
    """
    ctx.forward(cookbook_update)


if __name__ == "__main__":
    sys.argv[0] = "mussels"
    cli(sys.argv[1:])



"""
Copyright (C) 2019-2020 Cisco Systems, Inc. and/or its affiliates. All rights reserved.

A collections of helpers for Click

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import click

#
# Click command-group modifiers
#
class MusselsModifier(click.Group):
    def get_command(self, ctx, cmd_name):
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv
        matches = [x for x in self.list_commands(ctx) if x.startswith(cmd_name)]
        if not matches:
            return None
        elif len(matches) == 1:
            return click.Group.get_command(self, ctx, matches[0])
        ctx.fail("Too many matches: %s" % ", ".join(sorted(matches)))

    def format_epilog(self, ctx, formatter):
        if self.epilog:
            print(self.epilog)


class ShortNames(click.Group):
    def get_command(self, ctx, cmd_name):
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv
        matches = [x for x in self.list_commands(ctx) if x.startswith(cmd_name)]
        if not matches:
            return None
        elif len(matches) == 1:
            return click.Group.get_command(self, ctx, matches[0])
        ctx.fail("Too many matches: %s" % ", ".join(sorted(matches)))



"""
Copyright (C) 2019-2020 Cisco Systems, Inc. and/or its affiliates. All rights reserved.

This module provides the core Mussels class, used by the CLI interface defined in __main__.py

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from collections import defaultdict
from pathlib import Path

import datetime
import fnmatch
import json
import logging
import os
import platform
import shutil
import sys
import time
from typing import *

if platform.system() == "Windows":
    if not r"c:\program files\git\cmd" in os.environ["PATH"].lower():
        os.environ["PATH"] = os.environ["PATH"] + r";C:\Program Files\Git\cmd"
    if not r"c:\program files\git\mingw64\bin" in os.environ["PATH"].lower():
        os.environ["PATH"] = os.environ["PATH"] + r";C:\Program Files\Git\mingw64\bin"
    if not r"c:\program files\git\usr\bin" in os.environ["PATH"].lower():
        os.environ["PATH"] = os.environ["PATH"] + r";C:\Program Files\Git\usr\bin"
    if not r"c:\program files\git\bin" in os.environ["PATH"].lower():
        os.environ["PATH"] = os.environ["PATH"] + r";C:\Program Files\Git\bin"
import git
import yaml

import mussels.bookshelf
import mussels.recipe
import mussels.tool
from mussels.utils.versions import (
    NVC,
    nvc_str,
    sort_cookbook_by_version,
    version_keys,
    get_item_version,
    platform_is,
    platform_matches,
    pick_platform,
)


class Mussels:
    config: dict = {}
    cookbooks: defaultdict = defaultdict(dict)

    recipes: defaultdict = defaultdict(dict)
    sorted_recipes: dict = {}

    tools: defaultdict = defaultdict(dict)
    sorted_tools: dict = {}

    log_level: str

    def __init__(
        self,
        load_all_recipes: bool = False,
        data_dir: str = os.path.join(str(Path.home()), ".mussels"),
        install_dir: str = "",
        work_dir: str = "",
        log_dir: str = "",
        download_dir: str = "",
        log_level: str = "DEBUG",
    ) -> None:
        """
        Mussels class.

        Args:
            data_dir:   path where ClamAV should be installed.
            log_file:   path output log.
            log_level:  log level ("DEBUG", "INFO", "WARNING", "ERROR").
        """
        if log_dir != "":
            self.log_file = os.path.join(log_dir, "mussels.log")
        else:
            self.log_file = os.path.join(data_dir, "logs", "mussels.log")
        self.log_level = log_level
        self._init_logging(log_level)

        self.app_data_dir = data_dir
        if install_dir == "":
            self.install_dir = os.path.join(self.app_data_dir, "install")
            self.custom_install_dir = False
        else:
            self.install_dir = os.path.abspath(install_dir)
            self.custom_install_dir = True

        self.work_dir = "" if work_dir == "" else os.path.abspath(work_dir)
        self.log_dir = "" if log_dir == "" else os.path.abspath(log_dir)
        self.download_dir = "" if download_dir == "" else os.path.abspath(download_dir)

        self._load_config("cookbooks.json", self.cookbooks)
        self._load_recipes(all=load_all_recipes)

    def _init_logging(self, level="DEBUG"):
        """
        Initializes the logging parameters

        Returns:    nothing
        """
        levels = {
            "DEBUG": logging.DEBUG,
            "INFO": logging.INFO,
            "WARN": logging.WARNING,
            "WARNING": logging.WARNING,
            "ERROR": logging.ERROR,
        }

        self.logger = logging.getLogger("Mussels")
        self.logger.setLevel(levels[level])

        formatter = logging.Formatter(
            fmt="%(asctime)s - %(levelname)s:  %(message)s",
            datefmt="%m/%d/%Y %I:%M:%S %p",
        )

        if not os.path.exists(os.path.split(self.log_file)[0]):
            os.makedirs(os.path.split(self.log_file)[0])
        self.filehandler = logging.FileHandler(filename=self.log_file)
        self.filehandler.setLevel(levels[level])
        self.filehandler.setFormatter(formatter)

        self.logger.addHandler(self.filehandler)

    def _load_config(self, filename, config) -> bool:
        """
        Load the cache.
        """
        # load config, if exists.
        try:
            with open(
                os.path.join(self.app_data_dir, "config", filename), "r"
            ) as config_file:
                config.update(json.load(config_file))
        except Exception:
            # No existing config to load, that's probaby ok, but return false to indicate the failure.
            return False

        return True

    def _store_config(self, filename, config) -> bool:
        """
        Update the cache.
        """
        try:
            if not os.path.isdir(os.path.join(self.app_data_dir, "config")):
                os.makedirs(os.path.join(self.app_data_dir, "config"))
        except Exception as exc:
            self.logger.warning(f"Failed to create config directory.  Exception: {exc}")
            return False

        try:
            with open(
                os.path.join(self.app_data_dir, "config", filename), "w"
            ) as config_file:
                json.dump(config, config_file, indent=4)
        except Exception as exc:
            self.logger.warning(f"Failed to update config.  Exception: {exc}")
            return False

        return True

    def load_directory(self, cookbook: str, load_path: str) -> tuple:
        """
        Load all recipes and tools in a directory.
        This function reads in YAML files and assigns each to a new Recipe or Tool class, accordingly.
        The classes are returned in a tuple.
        """
        minimum_version = "0.1"
        recipes = defaultdict(dict)
        tools = defaultdict(dict)

        if not os.path.exists(load_path):
            return recipes, tools

        for root, dirs, filenames in os.walk(load_path):
            for fname in filenames:
                if not fname.endswith(".yaml"):
                    continue
                fpath = os.path.abspath(os.path.join(root, fname))
                with open(fpath, "r") as fd:
                    try:
                        yaml_file = yaml.load(fd.read(), Loader=yaml.SafeLoader)
                    except Exception as exc:
                        self.logger.warning(f"Failed to load YAML file: {fpath}")
                        self.logger.warning(f"Exception occured: \n{exc}")
                        continue
                    if yaml_file == None:
                        continue

                    if (
                        "mussels_version" in yaml_file
                        and yaml_file["mussels_version"] >= minimum_version
                    ):
                        if not "type" in yaml_file:
                            self.logger.warning(f"Failed to load recipe: {fpath}")
                            self.logger.warning(f"Missing required 'type' field.")
                            continue

                        if (
                            yaml_file["type"] == "recipe"
                            or yaml_file["type"] == "collection"
                        ):
                            if not "name" in yaml_file:
                                self.logger.warning(f"Failed to load recipe: {fpath}")
                                self.logger.warning(f"Missing required 'name' field.")
                                continue
                            name = f"{cookbook}__{yaml_file['name']}"

                            if not "version" in yaml_file:
                                self.logger.warning(f"Failed to load recipe: {fpath}")
                                self.logger.warning(
                                    f"Missing required 'version' field."
                                )
                                continue
                            else:
                                name = f"{name}_{yaml_file['version']}"

                            recipe_class = type(
                                name,
                                (mussels.recipe.BaseRecipe,),
                                {"__doc__": f"{yaml_file['name']} recipe class."},
                            )

                            recipe_class.module_file = fpath

                            recipe_class.name = yaml_file["name"]

                            recipe_class.version = yaml_file["version"]

                            if yaml_file["type"] == "collection":
                                recipe_class.is_collection = True
                            else:
                                recipe_class.is_collection = False

                                if not "url" in yaml_file:
                                    self.logger.warning(
                                        f"Failed to load recipe: {fpath}"
                                    )
                                    self.logger.warning(
                                        f"Missing required 'url' field."
                                    )
                                    continue
                                else:
                                    recipe_class.url = yaml_file["url"]

                            if "archive_name_change" in yaml_file:
                                recipe_class.archive_name_change = (
                                    yaml_file["archive_name_change"][0],
                                    yaml_file["archive_name_change"][1],
                                )

                            if not "platforms" in yaml_file:
                                self.logger.warning(f"Failed to load recipe: {fpath}")
                                self.logger.warning(
                                    f"Missing required 'platforms' field."
                                )
                                continue
                            else:
                                recipe_class.platforms = yaml_file["platforms"]

                            recipes[recipe_class.name][
                                recipe_class.version
                            ] = recipe_class

                        elif yaml_file["type"] == "tool":
                            if not "name" in yaml_file:
                                self.logger.warning(f"Failed to load tool: {fpath}")
                                self.logger.warning(f"Missing required 'name' field.")
                                continue
                            name = f"{cookbook}__{yaml_file['name']}"

                            if "version" in yaml_file:
                                name = f"{name}_{yaml_file['version']}"

                            tool_class = type(
                                name,
                                (mussels.tool.BaseTool,),
                                {"__doc__": f"{yaml_file['name']} tool class."},
                            )

                            tool_class.module_file = fpath

                            tool_class.name = yaml_file["name"]

                            if "version" in yaml_file:
                                tool_class.version = yaml_file["version"]

                            if not "platforms" in yaml_file:
                                self.logger.warning(f"Failed to load tool: {fpath}")
                                self.logger.warning(
                                    f"Missing required 'platforms' field."
                                )
                                continue
                            else:
                                tool_class.platforms = yaml_file["platforms"]

                            tools[tool_class.name][tool_class.version] = tool_class

        return recipes, tools

    def _read_cookbook(self, cookbook: str, cookbook_path: str) -> bool:
        """
        Load the recipes and tools from a single cookbook.
        """

        sorted_recipes: defaultdict = defaultdict(list)
        sorted_tools: defaultdict = defaultdict(list)

        # Load the recipes and the tools
        recipes, tools = self.load_directory(
            cookbook=cookbook, load_path=os.path.join(cookbook_path)
        )

        # Sort the recipes
        sorted_recipes = sort_cookbook_by_version(recipes)

        if len(sorted_recipes) > 0:
            self.cookbooks[cookbook]["recipes"] = sorted_recipes
            for recipe in recipes.keys():
                for version in recipes[recipe]:
                    if version not in self.recipes[recipe].keys():
                        self.recipes[recipe][version] = {}
                    self.recipes[recipe][version][cookbook] = recipes[recipe][version]

        # Sort the tools
        sorted_tools = sort_cookbook_by_version(tools)

        if len(sorted_tools) > 0:
            self.cookbooks[cookbook]["tools"] = sorted_tools
            for tool in tools.keys():
                for version in tools[tool]:
                    if version not in self.tools[tool].keys():
                        self.tools[tool][version] = {}
                    self.tools[tool][version][cookbook] = tools[tool][version]

        if len(recipes) == 0 and len(tools) == 0:
            return False

        if "trusted" not in self.cookbooks[cookbook]:
            self.cookbooks[cookbook]["trusted"] = False

        return True

    def _read_bookshelf(self) -> bool:
        """
        Load the recipes and tools from cookbooks in ~/.mussels/cookbooks
        """
        bookshelf = os.path.join(self.app_data_dir, "cookbooks")
        if os.path.isdir(bookshelf):
            for cookbook in os.listdir(bookshelf):
                cookbook_path = os.path.join(
                    os.path.join(self.app_data_dir, "cookbooks"), cookbook
                )
                if os.path.isdir(cookbook_path):
                    if not self._read_cookbook(cookbook, cookbook_path):
                        self.logger.warning(
                            f"Failed to read any recipes or tools from cookbook: {cookbook}"
                        )

            self._store_config("cookbooks.json", self.cookbooks)

        return True

    def _read_local_recipes(self) -> bool:
        """
        Load the recipes and tools from local "mussels" directory
        """
        # Load recipes and tools from `cwd` directory, if any exist.
        local_recipes = os.path.join(os.getcwd())
        if os.path.isdir(local_recipes):
            if not self._read_cookbook("local", local_recipes):
                return False

            self.cookbooks["local"]["url"] = ""
            self.cookbooks["local"]["path"] = local_recipes
            self.cookbooks["local"]["trusted"] = True

        return True

    def _sort_items_by_version(
        self, items: defaultdict, all: bool, has_target: bool = False
    ) -> dict:
        """
        Sort recipes, and determine the highest versions.
        Only includes trusted recipes for current platform.
        """
        sorted_items: dict = {}

        for item in items:
            versions_list = list(items[item].keys())
            versions_list.sort(key=version_keys)
            versions_list.reverse()

            sorted_item_list = []

            for version in versions_list:
                found_good_version = False
                item_version = {"version": version, "cookbooks": {}}

                for each_cookbook in items[item][version].keys():
                    if not all and not self.cookbooks[each_cookbook]["trusted"]:
                        continue

                    cookbook: dict = {}

                    for each_platform in items[item][version][each_cookbook].platforms:
                        if not all and not platform_is(each_platform):
                            continue

                        if has_target:
                            cookbook[each_platform] = [
                                target
                                for target in items[item][version][each_cookbook]
                                .platforms[each_platform]
                                .keys()
                            ]
                        else:
                            cookbook[each_platform] = []

                        found_good_version = True

                    item_version["cookbooks"][each_cookbook] = cookbook

                if found_good_version:
                    sorted_item_list.append(item_version)

            if len(sorted_item_list) > 0:
                sorted_items[item] = sorted_item_list

        return sorted_items

    def _load_recipes(self, all: bool = False) -> bool:
        """
        Load the recipes and tools.
        """
        # If the cache is empty, try reading from the local bookshelf.
        if len(self.recipes) == 0 or len(self.tools) == 0:
            self._read_bookshelf()

        # Load recipes from the local mussels directory, if those exists.
        if not self._read_local_recipes() and "local" in self.cookbooks:
            self.cookbooks.pop("local")

        if len(self.recipes) == 0:
            return False

        self.sorted_recipes = self._sort_items_by_version(
            self.recipes, all=all, has_target=True
        )
        self.sorted_tools = self._sort_items_by_version(self.tools, all=all)

        if len(self.sorted_recipes) == 0 or len(self.sorted_tools) == 0:
            return False

        return True

    def _build_recipe(
        self,
        recipe: str,
        version: str,
        cookbook: str,
        platform: str,
        target: str,
        toolchain: dict,
        rebuild: bool = False,
    ) -> dict:
        """
        Build a specific recipe.

        Args:
            recipe:     The recipe name with no version information.
            version:    The recipe version.

        Returns:    A dictionary of build results
        """
        result = {"name": recipe, "version": version, "success": False}

        if not self.cookbooks[cookbook]["trusted"]:
            self.logger.error(
                f"Unable to build {recipe}={version} from '{cookbook}'. You have not elected to trust '{cookbook}'"
            )
            self.logger.error(
                f"Building recipes involve downloading and executing code from the internet, which carries some risk."
            )
            self.logger.error(
                f"Please review the recipes provided by '{cookbook}' at: {self.cookbooks[cookbook]['url']}."
            )
            self.logger.error(
                f"If you're comfortable with the level of risk, run the following command to trust all recipes from '{cookbook}':"
            )
            self.logger.error(f"")
            self.logger.error(f"    mussels cookbook trust {cookbook}")
            self.logger.error(f"")
            self.logger.error(
                f"Alternatively, you may consider cloning only the recipe you need for your own cookbook."
            )
            self.logger.error(
                f"This is a safer option, though you are still encouraged to review the recipe before using it."
            )
            self.logger.error(
                f"To clone the recipe {recipe}={version} from '{cookbook}', run the following command:"
            )
            self.logger.error(f"")
            self.logger.error(
                f"    mussels recipe clone {recipe} -v {version} -c {cookbook}"
            )
            return result

        start = time.time()

        self.logger.info(f"Attempting to build {recipe}...")

        if version == "":
            # Use the default (highest) version
            try:
                version = self.sorted_recipes[recipe][0]
            except KeyError:
                self.logger.error(f"FAILED to find recipe: {recipe}!")
                result["time elapsed"] = time.time() - start
                return result

        try:
            recipe_class = self.recipes[recipe][version][cookbook]
        except KeyError:
            self.logger.error(f"FAILED to find recipe: {nvc_str(recipe, version)}!")
            result["time elapsed"] = time.time() - start
            return result

        # If the user specified a custom install directory, then don't add the target arch subdirectory.
        if self.custom_install_dir == True:
            install_dir = self.install_dir
        else:
            install_dir = os.path.join(self.install_dir, target)

        recipe_object = recipe_class(
            toolchain=toolchain,
            platform=platform,
            target=target,
            data_dir=self.app_data_dir,
            install_dir=install_dir,
            work_dir=self.work_dir,
            log_dir=self.log_dir,
            download_dir=self.download_dir,
            log_level=self.log_level,
        )

        if not recipe_object.build(rebuild):
            self.logger.error(f"FAILURE: {nvc_str(recipe, version)} build failed!\n")
        else:
            self.logger.info(
                f"Success: {nvc_str(recipe, version)} build succeeded. :)\n"
            )
            result["success"] = True

        result["time elapsed"] = time.time() - start

        return result

    def _get_recipe_version(self, recipe: str, platform: str, target: str) -> NVC:
        """
        Select recipe version based on version requirements.
        Eliminate recipe versions and sorted tools versions based on
        these requirements, and the required_tools requirements of remaining recipes.

        Args:
            recipe:     A specific recipe string, which may include version information.
            cookbook:   The preferred cookbook to select the recipe from.

        :return: named tuple describing the highest qualified version:
            NVC(
                "name"->str,
                "version"->str,
                "cookbook"->str,
            )
        """
        # Select the recipe
        nvc = get_item_version(recipe, self.sorted_recipes, target, logger=self.logger)

        # Use "get_item_version()" to prune the list of sorted_tools based on the required tools for the selected recipe.
        recipe_class = self.recipes[nvc.name][nvc.version][nvc.cookbook]

        for each_platform in recipe_class.platforms:
            if platform_matches(each_platform, platform):
                variant = recipe_class.platforms[each_platform]
                if target in variant.keys():
                    build_target = variant[target]

                    if "required_tools" in build_target.keys():
                        for tool in build_target["required_tools"]:
                            try:
                                get_item_version(tool, self.sorted_tools, logger=self.logger)
                            except Exception as exc:
                                raise Exception(f"The {tool} tool, required by {nvc_str(nvc.name, nvc.version, nvc.cookbook)} is not available...\n{exc}")
                    break
        return nvc

    def _identify_build_recipes(
        self, recipe: str, chain: list, platform: str, target: str
    ) -> list:
        """
        Identify all recipes that must be built given a specific recipe.

        Args:
            recipe:     A specific recipe to build.
            chain:      (in,out) A dependency chain starting from the first
                        recursive call used to identify circular dependencies.
        """
        recipe_nvc = self._get_recipe_version(recipe, platform, target)

        if (len(chain) > 0) and (recipe_nvc.name == chain[0]):
            raise ValueError(f"Circular dependencies found! {chain}")
        chain.append(recipe_nvc.name)

        recipes = []

        recipes.append(recipe)

        # Verify that recipe supports current platform.
        platform_options = self.recipes[recipe_nvc.name][recipe_nvc.version][
            recipe_nvc.cookbook
        ].platforms.keys()
        matching_platform = pick_platform(platform, platform_options)
        if matching_platform == "":
            # recipe doesn't support current platform.
            # TODO: see if next recipe does.
            pass

        # verify that recipe supports requested target architecture

        if (
            "dependencies"
            in self.recipes[recipe_nvc.name][recipe_nvc.version][
                recipe_nvc.cookbook
            ].platforms[matching_platform][target]
        ):
            dependencies = self.recipes[recipe_nvc.name][recipe_nvc.version][
                recipe_nvc.cookbook
            ].platforms[matching_platform][target]["dependencies"]
            for dependency in dependencies:
                if ":" not in dependency:
                    # If the cookbook isn't explicitly specified for the dependency,
                    # select the recipe from the current cookbook.
                    dependency = f"{recipe_nvc.cookbook}:{dependency}"

                try:
                    recipes += self._identify_build_recipes(
                        dependency, chain, platform, target
                    )
                except Exception as exc:
                    raise Exception(f"The {dependency} recipe, required by {nvc_str(recipe_nvc.name, recipe_nvc.version, recipe_nvc.cookbook)} has dependency issues...\n{exc}")

        return recipes

    def _get_build_batches(self, recipe: str, platform: str, target: str) -> list:
        """
        Get list of build batches that can be built concurrently.

        Args:
            recipe:    A recipes string in the format [cookbook:]recipe[==version].
        """
        # Identify all recipes that must be built given list of desired builds.
        try:
            all_recipes = set(self._identify_build_recipes(recipe, [], platform, target))
        except Exception as exc:
            raise Exception(f"Failed to assemble dependency chain for {recipe} on {platform} ({target}):\n{exc}")

        # Build a map of recipes (name,version) tuples to sets of dependency (name,version,cookbook) tuples
        nvc_to_deps = {}
        for recipe in all_recipes:
            recipe_nvc = self._get_recipe_version(recipe, platform, target)
            platform_options = self.recipes[recipe_nvc.name][recipe_nvc.version][
                recipe_nvc.cookbook
            ].platforms.keys()
            matching_platform = pick_platform(platform, platform_options)

            if (
                "dependencies"
                in self.recipes[recipe_nvc.name][recipe_nvc.version][
                    recipe_nvc.cookbook
                ].platforms[matching_platform][target]
            ):
                dependencies = self.recipes[recipe_nvc.name][recipe_nvc.version][
                    recipe_nvc.cookbook
                ].platforms[matching_platform][target]["dependencies"]
            else:
                dependencies = []
            nvc_to_deps[recipe_nvc] = set(
                [
                    self._get_recipe_version(dependency, platform, target)
                    for dependency in dependencies
                ]
            )

        batches = []

        # While there are dependencies to solve...
        while nvc_to_deps:

            # Get all recipes with no dependencies
            ready = {recipe for recipe, deps in nvc_to_deps.items() if not deps}

            # If there aren't any, we have a loop in the graph
            if not ready:
                msg = "Circular dependencies found!\n"
                msg += json.dumps(nvc_to_deps, indent=4)
                raise ValueError(msg)

            # Remove them from the dependency graph
            for recipe in ready:
                del nvc_to_deps[recipe]
            for deps in nvc_to_deps.values():
                deps.difference_update(ready)

            # Add the batch to the list
            batches.append(ready)

        # Return the list of batches
        return batches

    def _select_cookbook(
        self, recipe: str, recipe_version: dict, preferred_book: str = ""
    ) -> str:
        """
        Return the cookbook name, if only one cookbook provides the recipe-version.
        If more then one cookbook provides the recipe-version, explain the options and return an empty string.
        """
        cookbook = ""

        num_cookbooks = len(recipe_version["cookbooks"].keys())
        if num_cookbooks == 0:
            self.logger.error(
                f"Recipe {nvc_str(recipe, recipe_version['version'])} not provided by any cookbook!(?!)"
            )

        elif num_cookbooks == 1:
            cookbook = next(iter(recipe_version["cookbooks"]))

        else:
            if "local" in recipe_version["cookbooks"]:
                # Always prefer to use a local recipe.
                cookbook = "local"
            elif preferred_book != "" and preferred_book in recipe_version["cookbooks"]:
                # 2nd choice is the "preferred" cookbook, which was probably the same cookbook as the target recipe.
                cookbook = preferred_book
            else:
                # More than one option exists, but no good excuse to choose one over another.
                # Bail out and ask for more specific instructions.
                self.logger.error(
                    f'Failed to select a cookbook for {nvc_str(recipe, recipe_version["version"])}'
                )
                self.logger.error(
                    f"No cookbook specified, no local recipe exists, and no recipe exists in the same cookbook as the primary build target recipe."
                )
                self.logger.error(
                    f"However, multiple cookbooks do provide the recipe. Please retry with a specific cookbook using the `-c` / `--cookbook` options"
                )
                self.logger.info(f"")

                self.print_recipe_details(
                    recipe, recipe_version, verbose=True, all=True
                )

        return cookbook

    def check_tool(
        self,
        tool: str,
        version: str,
        cookbook: str,
        results: list,
    ) -> bool:
        """
        Check if a tool exists. Will check all tools if tool arg is "".

        Args:
            recipe:     The recipe to build.
            version:    A specific version to build.  Leave empty ("") to build the newest.
            cookbook:   A specific cookbook to use.  Leave empty ("") if there's probably only one.
            results:    (out) A list of dictionaries describing the results of the build.
        """
        found_tool = False

        for each_tool in self.sorted_tools:
            if tool == "" or tool == each_tool:
                for each_version in self.sorted_tools[each_tool]:
                    if version == "" or version == each_version["version"]:
                        for each_cookbook in each_version["cookbooks"]:
                            if cookbook == "" or cookbook == each_cookbook:
                                found_tool = True

                                tool_class = self.tools[each_tool][each_version["version"]][each_cookbook]
                                tool_object = tool_class(
                                    self.app_data_dir,
                                    log_level=self.log_level,
                                )

                                if tool_object.detect():
                                    # Found!
                                    self.logger.warning(
                                        f"    {nvc_str(each_tool, each_version['version'], each_cookbook)} FOUND."
                                    )
                                else:
                                    # Not found.
                                    self.logger.error(
                                        f"    {nvc_str(each_tool, each_version['version'], each_cookbook)} NOT found."
                                    )
        if not found_tool:
            self.logger.warning(
                f"    Unable to find tool definition matching: {nvc_str(tool, version, cookbook)}."
            )

    def build_recipe(
        self,
        recipe: str,
        version: str,
        cookbook: str,
        target: str,
        results: list,
        dry_run: bool = False,
        rebuild: bool = False,
    ) -> bool:
        """
        Execute a build of a recipe.

        Args:
            recipe:     The recipe to build.
            version:    A specific version to build.  Leave empty ("") to build the newest.
            cookbook:   A specific cookbook to use.  Leave empty ("") if there's probably only one.
            target:     The target architecture to build.
            results:    (out) A list of dictionaries describing the results of the build.
            dry_run:    (optional) Don't actually build, just print the build chain.
            rebuild:    (optional) Rebuild the entire dependency chain.
        """

        def print_results(results: list):
            """
            Print the build results in a pretty way.

            Args:
                results:    (out) A list of dictionaries describing the results of the build.
            """
            for result in results:
                if result["success"]:
                    self.logger.info(
                        f"Successful build of {nvc_str(result['name'], result['version'])} completed in {datetime.timedelta(0, result['time elapsed'])}."
                    )
                else:
                    self.logger.error(
                        f"Failure building {nvc_str(result['name'], result['version'])}, terminated after {datetime.timedelta(0, result['time elapsed'])}"
                    )

        if not recipe in self.sorted_recipes:
            self.logger.error(f"The recipe does not exist, or at least does not exist for the current platform ({platform.system()}")
            self.logger.error(f"To available recipes for your platform, run:   msl list")
            self.logger.error(f"To all recipes for all platforms, run:         msl list -a")
            self.logger.error(f"To download the latest recipes, run:           msl update")
            return False


        batches: List[dict] = []

        recipe_str = nvc_str(recipe, version, cookbook)

        if target == "":
            if platform.system() == "Windows":
                target = (
                    "x64" if os.environ["PROCESSOR_ARCHITECTURE"] == "AMD64" else "x86"
                )
            else:
                target = "host"

        try:
            batches = self._get_build_batches(
                recipe_str, platform=platform.system(), target=target
            )
        except Exception as exc:
            self.logger.error(f"{recipe_str} build failed!")
            for line in str(exc).split('\n'):
                self.logger.warning(f"{line}")
            return False

        #
        # Validate toolchain
        #
        # Collect set of required tools for entire build.
        toolchain = {}
        preferred_tool_versions = set()
        for i, bundle in enumerate(batches):
            for j, recipe_nvc in enumerate(bundle):
                recipe_class = self.recipes[recipe_nvc.name][recipe_nvc.version][
                    recipe_nvc.cookbook
                ]

                for each_platform in recipe_class.platforms:
                    if platform_is(each_platform):
                        if (
                            "required_tools"
                            in recipe_class.platforms[each_platform][target].keys()
                        ):
                            for tool in recipe_class.platforms[each_platform][target][
                                "required_tools"
                            ]:
                                tool_nvc = get_item_version(tool, self.sorted_tools, logger=self.logger)
                                preferred_tool_versions.add(tool_nvc)

        # Check if required tools are installed
        missing_tools = []
        for tool_nvc in preferred_tool_versions:
            tool_found = False
            preferred_tool = self.tools[tool_nvc.name][tool_nvc.version][
                tool_nvc.cookbook
            ](self.app_data_dir)

            if preferred_tool.detect():
                # Preferred tool version is available.
                tool_found = True
                toolchain[tool_nvc.name] = preferred_tool
                self.logger.info(
                    f"    {nvc_str(tool_nvc.name, tool_nvc.version, tool_nvc.cookbook)} found."
                )
            else:
                # Check if non-preferred (older, but compatible) version is available.
                self.logger.debug(
                    f"    {nvc_str(tool_nvc.name, tool_nvc.version, tool_nvc.cookbook)} not found."
                )

                if len(self.sorted_tools[tool_nvc.name]) > 1:
                    self.logger.debug(f"        Checking for alternative versions...")
                    alt_versions = self.sorted_tools[tool_nvc.name][1:]

                    for alt_version in alt_versions:
                        alt_version_cookbook = self._select_cookbook(
                            tool_nvc.name, alt_version, cookbook
                        )
                        alt_tool = self.tools[tool_nvc.name][alt_version["version"]][
                            alt_version_cookbook
                        ](self.app_data_dir)

                        if alt_tool.detect():
                            # Found a compatible version to use.
                            tool_found = True
                            toolchain[tool_nvc.name] = alt_tool

                            # Select the exact version (pruning all other options) so it will be the default.
                            get_item_version(
                                f"{nvc_str(tool_nvc.name, alt_version['version'], alt_version_cookbook)}",
                                self.sorted_tools,
                                logger=self.logger
                            )
                            self.logger.info(
                                f"    Alternative version {nvc_str(tool_nvc.name, alt_version['version'], alt_version_cookbook)} found."
                            )
                        else:
                            self.logger.debug(
                                f"    Alternative version {nvc_str(tool_nvc.name, alt_version['version'], alt_version_cookbook)} not found."
                            )

                if not tool_found:
                    # Tool is missing.  Build will fail.
                    missing_tools.append(tool_nvc)

        if len(missing_tools) > 0:
            self.logger.warning("")
            self.logger.warning(
                "The following tools are missing and must be installed for this build to continue:"
            )
            for tool_version in missing_tools:
                self.logger.warning(f"    {nvc_str(tool_version.name, tool_version.version)}")

            sys.exit(1)

        self.logger.info("Toolchain:")
        for tool in toolchain:
            self.logger.info(f"   {nvc_str(tool, toolchain[tool].version)}")

        #FF
        # Perform Build
        #
        if dry_run:
            self.logger.warning("")
            self.logger.warning(r"    ___   ___   _         ___   _     _    ")
            self.logger.warning(r"   | | \ | |_) \ \_/     | |_) | | | | |\ |")
            self.logger.warning(r"   |_|_/ |_| \  |_|      |_| \ \_\_/ |_| \|")
            self.logger.warning("")
            self.logger.info("Build-order of requested recipes:")

        idx = 0
        failure = False
        for i, bundle in enumerate(batches):
            for j, recipe_nvc in enumerate(bundle):
                idx += 1

                platform_options = self.recipes[recipe_nvc.name][recipe_nvc.version][
                    recipe_nvc.cookbook
                ].platforms.keys()
                matching_platform = pick_platform(platform.system(), platform_options)

                if dry_run:
                    self.logger.info(
                        f"   {idx:2} [{i}:{j:2}]: {nvc_str(recipe_nvc.name, recipe_nvc.version, recipe_nvc.cookbook)}"
                    )
                    if (
                        "required_tools"
                        in self.recipes[recipe_nvc.name][recipe_nvc.version][
                            recipe_nvc.cookbook
                        ].platforms[matching_platform][target]
                    ):
                        self.logger.debug(f"      Tool(s):")
                        for tool in self.recipes[recipe_nvc.name][recipe_nvc.version][
                            recipe_nvc.cookbook
                        ].platforms[matching_platform][target]["required_tools"]:
                            tool_nvc = get_item_version(tool, self.sorted_tools, logger=self.logger)
                            self.logger.debug(
                                f"        {nvc_str(tool_nvc.name, tool_nvc.version, tool_nvc.cookbook)}"
                            )
                        continue

                if failure:
                    self.logger.warning(
                        f"Skipping  {nvc_str(recipe_nvc.name, recipe_nvc.version, recipe_nvc.cookbook)} build due to prior failure."
                    )
                else:
                    result = self._build_recipe(
                        recipe_nvc.name,
                        recipe_nvc.version,
                        recipe_nvc.cookbook,
                        matching_platform,
                        target,
                        toolchain,
                        rebuild,
                    )
                    results.append(result)
                    if not result["success"]:
                        failure = True

        if not dry_run:
            print_results(results)

        if failure:
            return False
        return True

    def print_recipe_details(
        self, recipe: str, version: dict, verbose: bool, all: bool
    ):
        """
        Print recipe information.
        """
        version_num = version["version"]
        cookbooks = version["cookbooks"].keys()
        self.logger.info(f"    {nvc_str(recipe, version_num)};  provided by cookbook(s): {list(cookbooks)}")

        if verbose:
            self.logger.info("")
            for cookbook in cookbooks:
                self.logger.info(f"      Cookbook: {cookbook}")

                book_recipe = self.recipes[recipe][version_num][cookbook]

                if book_recipe.is_collection:
                    self.logger.info(f"        Collection: Yes")
                else:
                    self.logger.info(f"        Collection: No")

                self.logger.info(f"        Platforms:")
                for each_platform in book_recipe.platforms:
                    if all or platform_is(each_platform):
                        self.logger.info(f"          Host platform: {each_platform}")

                        variant = book_recipe.platforms[each_platform]
                        for arch in variant.keys():
                            self.logger.info(f"            Target architecture: {arch}")
                            self.logger.info(
                                f"              Dependencies:      {', '.join(variant[arch]['dependencies'])}"
                            )
                            self.logger.info(
                                f"              Required tools:    {', '.join(variant[arch]['required_tools'])}"
                            )

                        if not all:
                            break
            self.logger.info("")

    def show_recipe(self, recipe_match: str, version_match: str, verbose: bool = False):
        """
        Search recipes for a specific recipe and print recipe details.
        """

        found = False

        if version_match == "":
            self.logger.info(f'Searching for recipe matching name: "{recipe_match}"...')
        else:
            self.logger.info(
                f'Searching for recipe matching name: "{recipe_match}", version: "{version_match}"...'
            )
        # Attempt to match the recipe name
        for recipe in self.sorted_recipes:
            if fnmatch.fnmatch(recipe, recipe_match):
                if version_match == "":
                    found = True

                    # Show info for every version
                    for version in self.sorted_recipes[recipe]:
                        self.print_recipe_details(recipe, version, verbose, all)
                    break
                else:
                    # Attempt to match the version too
                    for version in self.sorted_recipes[recipe]:
                        if fnmatch.fnmatch(version, version_match):
                            found = True

                            self.print_recipe_details(recipe, version, verbose, all)
                            break
                    if found:
                        break
        if not found:
            if version_match == "":
                self.logger.warning(f'No recipe matching name: "{recipe_match}"')
            else:
                self.logger.warning(
                    f'No recipe matching name: "{recipe_match}", version: "{version_match}"'
                )

    def clone_recipe(self, recipe: str, version: str, cookbook: str, destination: str):
        """
        Search recipes for a specific recipe and copy the file to the CWD.
        """

        def get_cookbook(recipe: str, recipe_version: dict) -> str:
            """
            Return the cookbook name, if only one cookbook provides the recipe-version.
            If more then one cookbook provides the recipe-version, explain the options and return an empty string.
            """
            cookbook = ""

            num_cookbooks = len(recipe_version["cookbooks"].keys())
            if num_cookbooks == 0:
                self.logger.error(
                    f"Recipe {nvc_str(recipe, version)} not provided by any cookbook!(?!)"
                )

            elif num_cookbooks == 1:
                cookbook = next(iter(recipe_version["cookbooks"]))

            else:
                self.logger.error(
                    f'Clone failed: No cookbook specified, and multiple cookbooks provide recipe "{nvc_str(recipe, recipe_version["version"])}"'
                )
                self.logger.error(
                    f"Please retry with a specific cookbook using the `-c` or `--cookbook` option:"
                )
                self.logger.info(f"")

                self.print_recipe_details(
                    recipe, recipe_version, verbose=True, all=True
                )

            return cookbook

        found = False

        self.logger.info(
            f'Attempting to clone recipe: "{nvc_str(recipe, version, cookbook)}"...'
        )

        try:
            recipe_versions = self.sorted_recipes[recipe]
        except KeyError:
            self.logger.error(f'Clone failed: No such recipe "{recipe}"')
            return False

        # Identify highest available version, for future reference.
        highest_recipe_version = recipe_versions[0]

        #
        # Now repeat the above if/else logic to select the exact recipe requested.
        #

        if version == "":
            if cookbook == "":
                # neither version nor cookbook was specified.
                self.logger.info(
                    f"No version or cookbook specified, will select highest available version."
                )
                version = highest_recipe_version["version"]

                cookbook = get_cookbook(recipe, highest_recipe_version)

                if cookbook == "":
                    return False

            else:
                # cookbook specified, but version wasn't.
                self.logger.info(
                    f'No version specified, will select highest version provided by cookbook: "{cookbook}".'
                )

                selected_recipe_version = {}

                for recipe_version in recipe_versions:
                    if cookbook in recipe_version["cookbooks"].keys():
                        selected_recipe_version = recipe_version
                        break

                if selected_recipe_version == {}:
                    self.logger.error(
                        f'Clone failed: Requested recipe "{recipe}" could not be found in cookbook: "{cookbook}".'
                    )
                    return False

                version = selected_recipe_version["version"]

                if (
                    selected_recipe_version["version"]
                    != highest_recipe_version["version"]
                ):
                    self.logger.warning(
                        f'The version selected from cookbook "{cookbook}" is not the highest version available.'
                    )
                    self.logger.warning(
                        f"A newer version appears to be available from other sources:"
                    )
                    self.logger.info(f"")
                    self.print_recipe_details(
                        recipe, highest_recipe_version, verbose=True, all=True
                    )

        else:
            # version specified
            if cookbook == "":
                self.logger.info(
                    f"No cookbook specified, will select recipe only if version is provided by only one cookbook."
                )

                selected_recipe_version = {}

                for recipe_version in recipe_versions:
                    if version == recipe_version["version"]:

                        cookbook = get_cookbook(recipe, recipe_version)
                        break

                if cookbook == "":
                    return False

            else:
                # version and cookbook specified.
                pass

        if destination == "":
            destination = os.getcwd()

        try:
            recipe_class = self.recipes[recipe][version][cookbook]
        except KeyError:
            self.logger.error(
                f'Clone failed: Requested recipe "{nvc_str(recipe, version, cookbook)}" could not be found.'
            )
            return False

        recipe_basename = os.path.basename(recipe_class.module_file)
        clone_path = os.path.join(destination, recipe_basename)

        try:
            shutil.copyfile(
                recipe_class.module_file, os.path.join(destination, recipe_basename)
            )

            patch_dirs_copied: list = []
            for each_platform in recipe_class.platforms:
                for target in recipe_class.platforms[each_platform]:
                    if (
                        "patches" in recipe_class.platforms[each_platform][target]
                        and recipe_class.platforms[each_platform][target]["patches"] != ""
                    ):
                        patch_dir = os.path.join(
                            os.path.split(recipe_class.module_file)[0],
                            recipe_class.platforms[each_platform][target]["patches"],
                        )
                        if patch_dir in patch_dirs_copied:
                            # Already got this one,
                            continue

                        if not os.path.exists(patch_dir):
                            self.logger.warning(
                                f"Unable to clone referenced patch directory: {patch_dir}"
                            )
                            self.logger.warning(f"Directory does not exist.")
                        else:
                            patches_basename = os.path.basename(patch_dir)
                            shutil.copytree(
                                patch_dir, os.path.join(destination, patches_basename)
                            )
                        patch_dirs_copied.append(patch_dir)
        except Exception as exc:
            self.logger.error(f"Clone failed.  Exception: {exc}")
            return False

        self.logger.info(
            f'Successfully cloned recipe "{nvc_str(recipe, version, cookbook)}" to:'
        )
        self.logger.info(f"    {clone_path}")

        return True

    def list_recipes(self, verbose: bool = False):
        """
        Print out a list of all recipes and all collections.
        """
        has_collections = False

        if len(self.sorted_recipes) == 0:
            if len(self.cookbooks) > 0:
                self.logger.warning(f"No recipes available from trusted cookbooks.")
                self.logger.warning(f"Recipes from \"untrusted\" cookbooks are hidden by default.\n")
                self.logger.info(f"Run the this to view that which cookbooks are available:")
                self.logger.info(f"    mussels cookbook list -V\n")
                self.logger.info(f"Run this to view recipes & tools from a specific cookbook:")
                self.logger.info(f"    mussels cookbook show <name> -V\n")
                self.logger.info(f"To view ALL recipes, use:")
                self.logger.info(f"    mussels recipe list -a")
                return
            else:
                self.logger.warning(f"Failed to load any recipes.\n")
                self.logger.info(f"Re-run Mussels from a directory containing Mussels recipe & tool definitions,")
                self.logger.info(f" or use `mussels update` to download recipes from the public cookbooks.")
                return

        self.logger.info("Recipes:")
        for recipe in self.sorted_recipes:
            newest_version = self.sorted_recipes[recipe][0]["version"]
            cookbooks = list(self.recipes[recipe][newest_version].keys())
            if not self.recipes[recipe][newest_version][cookbooks[0]].is_collection:
                if not verbose:
                    outline = f"    {recipe:10} "
                    for i, version in enumerate(self.sorted_recipes[recipe]):
                        if i == 0:
                            outline += f" {version['version']}"
                            if len(self.sorted_recipes[recipe]) > 1:
                                outline += "*"
                        else:
                            outline += f", {version['version']}"
                    outline += ""
                    self.logger.info(outline)
                else:
                    outline = f"    {recipe:10} "
                    for i, version in enumerate(self.sorted_recipes[recipe]):
                        if i == 0:
                            outline += f" {version['version']} {version['cookbooks']}"
                            if len(self.sorted_recipes[recipe]) > 1:
                                outline += "*"
                        else:
                            outline += f", {version['version']} {version['cookbooks']}"
                    outline += ""
                    self.logger.info(outline)

        for recipe in self.sorted_recipes:
            newest_version = self.sorted_recipes[recipe][0]["version"]
            cookbooks = list(self.recipes[recipe][newest_version].keys())
            if self.recipes[recipe][newest_version][cookbooks[0]].is_collection:
                if not has_collections:
                    self.logger.info("")
                    self.logger.info("Collections:")
                    has_collections = True

                if not verbose:
                    outline = f"    {recipe:10} "
                    for i, version in enumerate(self.sorted_recipes[recipe]):
                        if i == 0:
                            outline += f" {version['version']}"
                            if len(self.sorted_recipes[recipe]) > 1:
                                outline += "*"
                        else:
                            outline += f", {version['version']}"
                    outline += ""
                    self.logger.info(outline)
                else:
                    outline = f"    {recipe:10} "
                    for i, version in enumerate(self.sorted_recipes[recipe]):
                        if i == 0:
                            outline += f" {version['version']} {version['cookbooks']}"
                            if len(self.sorted_recipes[recipe]) > 1:
                                outline += "*"
                        else:
                            outline += f", {version['version']} {version['cookbooks']}"
                    outline += ""
                    self.logger.info(outline)

    def print_tool_details(
        self, tool: str, version: dict, verbose: bool, all: bool
    ):
        """
        Print tool information.
        """
        version_num = version["version"]
        cookbooks = version["cookbooks"].keys()
        self.logger.info(f"    {nvc_str(tool, version_num)};  provided by cookbook(s): {list(cookbooks)}")

        if verbose:
            self.logger.info("")
            for cookbook in cookbooks:
                self.logger.info(f"      Cookbook: {cookbook}")

                book_tool = self.tools[tool][version_num][cookbook]

                self.logger.info(f"        Platforms:")
                for each_platform in book_tool.platforms:
                    if all or platform_is(each_platform):
                        self.logger.info(f"          Host platform: {each_platform}")

                        details = yaml.dump(book_tool.platforms[each_platform], indent=4)
                        for detail in details.split('\n'):
                            self.logger.info("            " + detail)

                        if not all:
                            break
            self.logger.info("")

    def show_tool(self, tool_match: str, version_match: str, verbose: bool = False):
        """
        Search tools for a specific tool and print tool details.
        """

        found = False

        if version_match == "":
            self.logger.info(f'Searching for tool matching name: "{tool_match}"...')
        else:
            self.logger.info(
                f'Searching for tool matching name: "{tool_match}", version: "{version_match}"...'
            )
        # Attempt to match the tool name
        for tool in self.sorted_tools:
            if fnmatch.fnmatch(tool, tool_match):
                if version_match == "":
                    found = True

                    # Show info for every version
                    for version in self.sorted_tools[tool]:
                        self.print_tool_details(tool, version, verbose, all)
                    break
                else:
                    # Attempt to match the version too
                    for version in self.sorted_tools[tool]:
                        if fnmatch.fnmatch(version, version_match):
                            found = True

                            self.print_tool_details(tool, version, verbose, all)
                            break
                    if found:
                        break
        if not found:
            if version_match == "":
                self.logger.warning(f'No tool matching name: "{tool_match}"')
            else:
                self.logger.warning(
                    f'No tool matching name: "{tool_match}", version: "{version_match}"'
                )

    def clone_tool(self, tool: str, version: str, cookbook: str, destination: str):
        """
        Search tools for a specific tool and copy the file to the CWD.
        """

        def get_cookbook(tool: str, tool_version: dict) -> str:
            """
            Return the cookbook name, if only one cookbook provides the tool-version.
            If more then one cookbook provides the tool-version, explain the options and return an empty string.
            """
            cookbook = ""

            num_cookbooks = len(tool_version["cookbooks"].keys())
            if num_cookbooks == 0:
                self.logger.error(
                    f"tool {nvc_str(tool, version)} not provided by any cookbook!(?!)"
                )

            elif num_cookbooks == 1:
                cookbook = next(iter(tool_version["cookbooks"]))

            else:
                self.logger.error(
                    f'Clone failed: No cookbook specified, and multiple cookbooks provide tool "{nvc_str(tool, tool_version["version"])}"'
                )
                self.logger.error(
                    f"Please retry with a specific cookbook using the `-c` or `--cookbook` option:"
                )
                self.logger.info(f"")

                self.print_tool_details(
                    tool, tool_version, verbose=True, all=True
                )

            return cookbook

        found = False

        self.logger.info(
            f'Attempting to clone tool: "{nvc_str(tool, version, cookbook)}"...'
        )

        try:
            tool_versions = self.sorted_tools[tool]
        except KeyError:
            self.logger.error(f'Clone failed: No such tool "{tool}"')
            return False

        # Identify highest available version, for future reference.
        highest_tool_version = tool_versions[0]

        #
        # Now repeat the above if/else logic to select the exact tool requested.
        #

        if version == "":
            if cookbook == "":
                # neither version nor cookbook was specified.
                self.logger.info(
                    f"No version or cookbook specified, will select highest available version."
                )
                version = highest_tool_version["version"]

                cookbook = get_cookbook(version, highest_tool_version)

                if cookbook == "":
                    return False

            else:
                # cookbook specified, but version wasn't.
                self.logger.info(
                    f'No version specified, will select highest version provided by cookbook: "{cookbook}".'
                )

                selected_tool_version = {}

                for tool_version in tool_versions:
                    if cookbook in tool_version["cookbooks"].keys():
                        selected_tool_version = tool_version
                        break

                if selected_tool_version == {}:
                    self.logger.error(
                        f'Clone failed: Requested tool "{tool}" could not be found in cookbook: "{cookbook}".'
                    )
                    return False

                version = selected_tool_version["version"]

                if (
                    selected_tool_version["version"]
                    != highest_tool_version["version"]
                ):
                    self.logger.warning(
                        f'The version selected from cookbook "{cookbook}" is not the highest version available.'
                    )
                    self.logger.warning(
                        f"A newer version appears to be available from other sources:"
                    )
                    self.logger.info(f"")
                    self.print_tool_details(
                        tool, highest_tool_version, verbose=True, all=True
                    )

        else:
            # version specified
            if cookbook == "":
                self.logger.info(
                    f"No cookbook specified, will select tool only if version is provided by only one cookbook."
                )

                selected_tool_version = {}

                for tool_version in tool_versions:
                    if version == tool_version["version"]:

                        cookbook = get_cookbook(tool, tool_version)
                        break

                if cookbook == "":
                    return False

            else:
                # version and cookbook specified.
                pass

        if destination == "":
            destination = os.getcwd()

        try:
            tool_class = self.tools[tool][version][cookbook]
        except KeyError:
            self.logger.error(
                f'Clone failed: Requested tool "{nvc_str(tool, version, cookbook)}" could not be found.'
            )
            return False

        tool_basename = os.path.basename(tool_class.module_file)
        clone_path = os.path.join(destination, tool_basename)

        try:
            shutil.copyfile(
                tool_class.module_file, os.path.join(destination, tool_basename)
            )
        except Exception as exc:
            self.logger.error(f"Clone failed.  Exception: {exc}")
            return False

        self.logger.info(
            f'Successfully cloned tool "{nvc_str(tool, version, cookbook)}" to:'
        )
        self.logger.info(f"    {clone_path}")

        return True

    def list_tools(self, verbose: bool = False):
        """
        Print out a list of all tools and all collections.
        """
        has_collections = False

        if len(self.sorted_tools) == 0:
            if len(self.cookbooks) > 0:
                self.logger.warning(f"No tools available from trusted cookbooks.")
                self.logger.warning(f"Tools from \"untrusted\" cookbooks are hidden by default.\n")
                self.logger.info(f"Run the this to view that which cookbooks are available:")
                self.logger.info(f"    mussels cookbook list -V\n")
                self.logger.info(f"Run this to view recipes & tools from a specific cookbook:")
                self.logger.info(f"    mussels cookbook show <name> -V\n")
                self.logger.info(f"To view ALL tools, use:")
                self.logger.info(f"    mussels tool list -a")
                return
            else:
                self.logger.warning(f"Failed to load any tools.\n")
                self.logger.info(f"Re-run Mussels from a directory containing Mussels recipe & tool definitions,")
                self.logger.info(f" or use `mussels update` to download recipes from the public cookbooks.")
                return

        self.logger.info("Tools:")
        for tool in self.sorted_tools:
            newest_version = self.sorted_tools[tool][0]["version"]
            cookbooks = list(self.tools[tool][newest_version].keys())

            if not verbose:
                outline = f"    {tool:10} "
                for i, version in enumerate(self.sorted_tools[tool]):
                    if i == 0:
                        outline += f" {version['version']}"
                        if len(self.sorted_tools[tool]) > 1:
                            outline += "*"
                    else:
                        outline += f", {version['version']}"
                outline += ""
                self.logger.info(outline)
            else:
                outline = f"    {tool:10} "
                for i, version in enumerate(self.sorted_tools[tool]):
                    if i == 0:
                        outline += f" {version['version']} {version['cookbooks']}"
                        if len(self.sorted_tools[tool]) > 1:
                            outline += "*"
                    else:
                        outline += f", {version['version']} {version['cookbooks']}"
                outline += ""
                self.logger.info(outline)

    def update_cookbooks(self) -> None:
        """
        Attempt to update each cookbook in using Git to clone or pull each repo.
        If git isn't available, warn the user they should probably install Git and add it to their PATH.
        """
        # Create ~/.mussels/bookshelf if it doesn't already exist.
        os.makedirs(os.path.join(self.app_data_dir, "cookbooks"), exist_ok=True)

        # Get url for each cookbook from the mussels bookshelf.
        for book in mussels.bookshelf.cookbooks:
            repo_dir = os.path.join(self.app_data_dir, "cookbooks", book)
            self.cookbooks[book]["path"] = repo_dir
            self.cookbooks[book]["url"] = mussels.bookshelf.cookbooks[book]["url"]
            if "trusted" not in self.cookbooks[book]:
                self.cookbooks[book]["trusted"] = False

        for book in self.cookbooks:
            repo_dir = os.path.join(self.app_data_dir, "cookbooks", book)

            if "url" in self.cookbooks[book] and self.cookbooks[book]["url"] != "":
                if not os.path.isdir(repo_dir):
                    repo = git.Repo.clone_from(self.cookbooks[book]["url"], repo_dir)
                else:
                    repo = git.Repo(repo_dir)
                    repo.git.pull()

            self._read_cookbook(book, repo_dir)

        self._store_config("cookbooks.json", self.cookbooks)

    def list_cookbooks(self, verbose: bool = False):
        """
        Print out a list of all cookbooks.
        """

        if len(self.cookbooks) == 0:
            self.logger.warning(f"Failed to load any cookbooks.\n")
            self.logger.info(f"Re-run Mussels from a directory containing Mussels recipe & tool definitions,")
            self.logger.info(f" or use `mussels update` to download recipes from the public cookbooks.")
            return

        self.logger.info("Cookbooks:")
        for cookbook in self.cookbooks:
            self.logger.info(f"    {cookbook}")

            if verbose:
                if cookbook == "local":
                    self.logger.info(f"        url:     n/a")
                else:
                    self.logger.info(
                        f"        url:     {self.cookbooks[cookbook]['url']}"
                    )
                self.logger.info(f"        path:    {self.cookbooks[cookbook]['path']}")
                self.logger.info(
                    f"        trusted: {self.cookbooks[cookbook]['trusted']}"
                )
                self.logger.info(f"")

    def show_cookbook(self, cookbook_match: str, verbose: bool):
        """
        Search cookbooks for a specific cookbook and print the details.
        """
        found = False

        self.logger.info(f'Searching for cookbook matching name: "{cookbook_match}"...')

        # Attempt to match the cookbook name
        for cookbook in self.cookbooks:
            if fnmatch.fnmatch(cookbook, cookbook_match):
                found = True

                self.logger.info(f"    {cookbook}")
                if cookbook == "local":
                    self.logger.info(f"        url:     n/a")
                else:
                    self.logger.info(
                        f"        url:     {self.cookbooks[cookbook]['url']}"
                    )
                self.logger.info(f"        path:    {self.cookbooks[cookbook]['path']}")
                self.logger.info(
                    f"        trusted: {self.cookbooks[cookbook]['trusted']}"
                )

                if verbose:
                    self.logger.info(f"")
                    if len(self.cookbooks[cookbook]["recipes"].keys()) > 0:
                        self.logger.info(f"    Recipes:")
                        for recipe in self.cookbooks[cookbook]["recipes"]:
                            self.logger.info(
                                f"        {recipe} : {self.cookbooks[cookbook]['recipes'][recipe]}"
                            )
                        self.logger.info(f"")
                    if len(self.cookbooks[cookbook]["tools"].keys()) > 0:
                        self.logger.info(f"    Tools:")
                        for tool in self.cookbooks[cookbook]["tools"]:
                            self.logger.info(
                                f"        {tool} : {self.cookbooks[cookbook]['tools'][tool]}"
                            )

        if not found:
            self.logger.warning(f'No cookbook matching name: "{cookbook_match}"')

    def clean_cache(self):
        """
        Clear the cache files.
        """
        self.logger.info(
            f"Clearing cache directory ( {os.path.join(self.app_data_dir, 'cache')} )..."
        )

        if os.path.exists(os.path.join(self.app_data_dir, "cache")):
            shutil.rmtree(os.path.join(self.app_data_dir, "cache"))
            self.logger.info(f"Cache directory cleared.")
        else:
            self.logger.info(f"No cache directory to clear.")

    def clean_install(self):
        """
        Clear the install files.
        """
        self.logger.info(
            f"Clearing install directory ( {os.path.join(self.app_data_dir, 'install')} )..."
        )

        if os.path.exists(os.path.join(self.install_dir)):
            shutil.rmtree(os.path.join(self.install_dir))
            self.logger.info(f"Install directory cleared.")
        else:
            self.logger.info(f"No install directory to clear.")

    def clean_logs(self):
        """
        Clear the log files.
        """
        self.logger.info(
            f"Clearing logs directory ( {os.path.join(self.app_data_dir, 'logs')} )..."
        )

        self.filehandler.close()
        self.logger.removeHandler(self.filehandler)

        if os.path.exists(os.path.join(self.app_data_dir, "logs")):
            shutil.rmtree(os.path.join(self.app_data_dir, "logs"))
            self.logger.info(f"Logs directory cleared.")
        else:
            self.logger.info(f"No logs directory to clear.")

    def clean_all(self):
        """
        Clear all Mussels files.
        """
        self.clean_cache()
        self.clean_install()
        self.clean_logs()

        self.logger.info(
            f"Clearing Mussels directory ( {os.path.join(self.app_data_dir)} )..."
        )

        if os.path.exists(os.path.join(self.app_data_dir)):
            shutil.rmtree(os.path.join(self.app_data_dir))
            self.logger.info(f"Mussels directory cleared.")
        else:
            self.logger.info(f"No Mussels directory to clear.")

    def config_trust_cookbook(self, cookbook):
        """
        Update config to indicate that a given cookbook is trusted.
        """
        if cookbook not in self.cookbooks:
            self.logger.error(
                f"Can't trust cookbook '{cookbook}'. Cookbook is unknown."
            )

        self.logger.info(f"'{cookbook}' cookbook is now trusted.")

        self.cookbooks[cookbook]["trusted"] = True

        self._store_config("cookbooks.json", self.cookbooks)

    def config_add_cookbook(self, cookbook, author, url, trust=False):
        """
        Update config to indicate that a given cookbook is trusted.
        """
        self.cookbooks[cookbook]["author"] = author
        self.cookbooks[cookbook]["url"] = url
        self.cookbooks[cookbook]["trusted"] = trust

        self._store_config("cookbooks.json", self.cookbooks)

    def config_remove_cookbook(self, cookbook):
        self.cookbooks.pop(cookbook)

        self._store_config("cookbooks.json", self.cookbooks)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="mussels",
    version="0.4.2",
    author="Micah Snyder",
    author_email="micasnyd@cisco.com",
    copyright="Copyright (C) 2025 Cisco Systems, Inc. and/or its affiliates. All rights reserved.",
    description="Mussels Dependency Build Automation Tool",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Cisco-Talos/mussels",
    packages=setuptools.find_packages(),
    entry_points={
        "console_scripts": [
            "mussels = mussels.__main__:cli",
            "msl = mussels.__main__:cli",
        ]
    },
    install_requires=[
        "click>=7.0",
        "coloredlogs>=10.0",
        "colorama",
        "requests",
        "patch",
        "gitpython",
        "pyyaml",
        "setuptools",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
)



"""
Copyright (C) 2019-2020 Cisco Systems, Inc. and/or its affiliates. All rights reserved.

This module provides the base class for every Tool.  A tool is anything that a Recipe
might depend on from the system environment.

The base class provides the logic for detecting tools.

There is work-in-progress logic to [optionally] download and install missing tools
on behalf of the user.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import datetime
from distutils import dir_util, spawn
import inspect
import logging
import os
import platform
import requests
import shutil
import stat
import subprocess
import tarfile
import zipfile

from io import StringIO

from mussels.utils.versions import platform_is, nvc_str


class BaseTool(object):
    """
    Base class for Mussels tool detection.
    """

    name = "sample"
    version = ""
    platforms: dict = {}
    logs_dir = ""
    tool_path = ""

    def __init__(self,
        data_dir: str = "",
        log_level: str = "DEBUG",
    ):
        """
        Download the archive (if necessary) to the Downloads directory.
        Extract the archive to the temp directory so it is ready to build.
        """
        if data_dir == "":
            # No temp dir provided, build in the current working directory.
            self.logs_dir = os.path.join(os.getcwd(), "logs", "tools")
        else:
            self.logs_dir = os.path.join(os.path.abspath(data_dir), "logs", "tools")
        os.makedirs(self.logs_dir, exist_ok=True)

        self.name_version = nvc_str(self.name, self.version)

        self._init_logging(log_level)

    def _init_logging(self, level="DEBUG"):
        """
        Initializes the logging parameters
        """
        levels = {
            "DEBUG": logging.DEBUG,
            "INFO": logging.INFO,
            "WARN": logging.WARNING,
            "WARNING": logging.WARNING,
            "ERROR": logging.ERROR,
        }

        self.logger = logging.getLogger(f"{self.name_version}")

        formatter = logging.Formatter(
            fmt="%(asctime)s - %(levelname)s:  %(message)s",
            datefmt="%m/%d/%Y %I:%M:%S %p",
        )

        self.log_file = os.path.join(
            self.logs_dir,
            f"{self.name_version}.{datetime.datetime.now()}.log".replace(":", "_"),
        )
        filehandler = logging.FileHandler(filename=self.log_file)
        filehandler.setLevel(logging.DEBUG)
        filehandler.setFormatter(formatter)

        self.logger.addHandler(filehandler)
        self.logger.setLevel(levels[os.environ.get("LOG_LEVEL", level)])

    def _run_command(self, command: str, expected_output: str) -> bool:
        """
        Run a command.
        """
        found_expected_output = False

        cmd = command.split()

        # Run the build script.
        try:
            process = subprocess.Popen(
                cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
            )
            with process.stdout:
                for line in iter(process.stdout.readline, b""):
                    if expected_output in line.decode("utf-8", errors='replace'):
                        found_expected_output = True

            process.wait()
            if process.returncode != 0:
                self.logger.debug(f"Command failed!")
                return False
        except FileNotFoundError:
            self.logger.debug(f"Command failed; File not found!")
            return False

        return found_expected_output

    def detect(self) -> bool:
        """
        Determine if tool is available in expected locations.
        """
        found = False

        self.logger.info(f"Detecting tool: {self.name_version}...")

        for each_platform in self.platforms:
            if platform_is(each_platform):
                if "path_checks" in self.platforms[each_platform]:
                    for path_check in self.platforms[each_platform]["path_checks"]:
                        self.logger.info(f"  Checking for {path_check} in PATH")
                        install_location = spawn.find_executable(path_check)
                        if install_location == None:
                            self.logger.info(f"    {path_check} not found")
                        else:
                            self.logger.info(
                                f"    {path_check} found, at: {install_location}"
                            )
                            found = True
                            break

                if found:
                    break

                if "command_checks" in self.platforms[each_platform]:
                    for script_check in self.platforms[each_platform]["command_checks"]:
                        found = self._run_command(
                            command=script_check["command"],
                            expected_output=script_check["output_has"],
                        )
                        if not found:
                            self.logger.info(f"    {script_check['command']} failed.")
                        else:
                            self.logger.info(f"    {script_check['command']} passed!")
                            break

                if found:
                    break

                if "file_checks" in self.platforms[each_platform]:
                    for filepath in self.platforms[each_platform]["file_checks"]:
                        if not os.path.exists(filepath):
                            self.logger.info(
                                f'{self.name_version} file "{filepath}" not found'
                            )
                        else:
                            self.logger.info(
                                f'{self.name_version} file "{filepath}" found'
                            )
                            self.tool_path = os.path.dirname(filepath)
                            found = True
                            break
                break

        if not found:
            self.logger.debug(f"Failed to detect {self.name_version}.")
            return False

        return True
